import weakref

from ..utils.trackable import Trackable


STRUCT_FORMAT_ALIASES = {"c": "B", "l": "i", "L": "I"}

FORMAT_MAP = {
    "b": "i1",
    "B": "u1",
    "h": "i2",
    "H": "u2",
    "i": "i4",
    "I": "u4",
    "e": "f2",
    "f": "f4",
}


def get_item_format_from_memoryview(mem):
    """Get the per-item format specifier from a memoryview.
    Returns None if the format appears to be a structured array.
    Raises an error if GPU-incompatible dtypes are used (64 bit).
    """

    # Uniform buffers are scalars with a structured dtype.
    # But can also create storage buffers with complex formats.
    if len(mem.format) > 2:
        return None

    # GPUs generally don't support 64-bit buffers or textures.
    # Note: the Python docs say that l and L are 32 bit, but converting
    # a int64 numpy array to a memoryview gives a format of 'l' instead
    # of 'q' on some systems/configs? So we need to check the itemsize.
    if mem.itemsize == 8:
        kind = "64-bit"
        if mem.format in "fd":
            kind = "float64"
        elif mem.format in "ilq":
            kind = "int64"
        elif mem.format in "ILQ":
            kind = "uint64"
        raise ValueError(
            f"A dtype of {kind} is not supported for buffers, use a 32-bit variant instead."
        )

    # Get normalized format
    format = str(mem.format)
    format = STRUCT_FORMAT_ALIASES.get(format, format)

    if format not in FORMAT_MAP:
        raise TypeError(
            f"Cannot convert {format!r} to wgpu format. You should provide data with a different dtype."
        )
    return FORMAT_MAP[format]


class Resource(Trackable):
    """Resource base class."""

    _resource_counts = {}
    _rev = 0

    def __init__(self):
        super().__init__()
        cname = self.__class__.__name__
        Resource._resource_counts[cname] = Resource._resource_counts.get(cname, 0) + 1

    def __del__(self):
        cname = self.__class__.__name__
        Resource._resource_counts[cname] -= 1

    def _gfx_mark_for_sync(self):
        resource_update_registry._gfx_mark_for_sync(self)


class ResourceUpdateRegistry:
    """Singleton registry to keep track of resources that need to be updated."""

    def __init__(self):
        self._syncable = weakref.WeakSet()

    def _gfx_mark_for_sync(self, resource):
        """Register the given resource for synchonization. Only adds the resource
        if it's wgpu-counterpart already exists, and when it has pending uploads.
        """
        # Note: this method is very specific to the wgpu renderer, and
        # its logic to handle updates to buffers and textures. We could
        # create some sort of plugin system so that each renderer can
        # register a registry (and that registry could attach the
        # _wgpu_xx attributes to the buffers/texture) but that feels
        # like overdesigning at this point. Let's track in #272.
        if not isinstance(resource, Resource):
            raise TypeError("Given object is not a Resource")
        if resource._wgpu_object is not None and resource._gfx_pending_uploads:
            self._syncable.add(resource)

    def get_syncable_resources(self, *, flush=False):
        """Get the set of resources that need syncing. If setting flush
        to True, the caller is responsible for syncing the resources.
        """
        syncable = set(self._syncable)
        if flush:
            self._syncable.clear()
        return syncable


resource_update_registry = ResourceUpdateRegistry()



import numpy as np

from ._base import Resource, get_item_format_from_memoryview


class Texture(Resource):
    """Texture object containing structured 1D, 2D or 3D data.

    Can be used to represent e.g. image data or colormaps. Can also
    serve as a render target (for the renderer). Supports texture
    stacks, cube textures, and mipmapping.

    Parameters:
        data : array, optional
            Array data of any type that supports the buffer-protocol, (e.g. a
            bytes or numpy array). If None, nbytes and nitems must be provided.
            The dtype must be compatible with the rendering backend.
        dim : int
            The dimensionality of the array (1, 2 or 3).
        size : tuple, [3]
            The extent ``(width, height, depth)`` of the array. If None, it is
            derived from `dim` and the shape of the data. The texture can also
            represent a stack of images by setting `dim=2` and `depth > 1`,
            or a cube image by setting `dim=2` and `depth==6`.
        format : str
            A format string describing the texture layout. If None, this is
            derived from the data. This must be a pygfx format
            specifier, e.g. "3xf4", but can also be a format specific to the
            render backend (e.g. from ``wgpu.TextureFormat``).
        colorspace : str
            If this data is used as color, it is interpreted to be in this
            colorspace. Can be "srgb" or "physical". Default "srgb".
        generate_mipmaps : bool
            If True, automatically generates mipmaps when transfering data to
            the GPU.

    """

    def __init__(
        self,
        data=None,
        *,
        dim,
        size=None,
        format=None,
        colorspace="srgb",
        generate_mipmaps=False,
    ):
        super().__init__()
        Resource._rev += 1
        self._rev = Resource._rev
        # The dim specifies the texture dimension
        assert dim in (1, 2, 3)
        self._store.dim = int(dim)
        # The actual data (optional)
        self._data = None
        self._format = None
        self._gfx_pending_uploads = []  # list of (offset, size) tuples

        # Backends-specific attributes for internal use
        self._wgpu_object = None
        self._wgpu_usage = 0
        self._wgpu_mip_level_count = 1

        self._colorspace = (colorspace or "srgb").lower()
        assert self._colorspace in ("srgb", "physical")

        self._generate_mipmaps = bool(generate_mipmaps)

        size = None if size is None else (int(size[0]), int(size[1]), int(size[2]))

        if data is not None:
            self._data = data
            self._mem = mem = memoryview(data)
            self._store.nbytes = mem.nbytes
            self._store.size = self._size_from_data(mem, dim, size)
            subformat = get_item_format_from_memoryview(mem)
            if subformat is None:
                raise ValueError(
                    f"Unsupported dtype/format for texture data: {mem.format}"
                )
            shape = mem.shape
            collapsed_size = [x for x in self.size if x > 1]
            if len(shape) == len(collapsed_size) + 1:
                nchannels = shape[-1]
            else:
                if not len(shape) == len(collapsed_size):
                    raise ValueError(
                        "Incompatible data shape for image data, there must be > 1 pixel to draw per channel"
                    )
                nchannels = 1
            if not (1 <= nchannels <= 4):
                raise ValueError(
                    f"Expected 1-4 texture color channels, got {nchannels}."
                )
            self._format = (f"{nchannels}x" + subformat).lstrip("1x")
            self.update_range((0, 0, 0), self.size)
        elif size is not None and format is not None:
            self._store.size = size
            self._store.nbytes = 0
        else:
            raise ValueError(
                "Texture must be instantiated with either data or size and format."
            )

        if format is not None:
            self._format = str(format)

    @property
    def rev(self):
        """An integer that is increased when update_range() is called."""
        return self._rev

    @property
    def dim(self):
        """The dimensionality of the texture (1, 2, or 3)."""
        return self._store.dim

    @property
    def data(self):
        """The data for this texture. Can be None if the data only
        exists on the GPU.

        Note: the data is the same reference that was given to
        instantiate this object, but this may change.
        """
        return self._data

    @property
    def mem(self):
        """The data for this buffer as a memoryview. Can be None if
        the data only exists on the GPU.
        """
        return self._mem

    @property
    def nbytes(self):
        """Get the number of bytes in the texture."""
        return self._store.nbytes

    @property
    def size(self):
        """The size of the texture as (width, height, depth).
        (always a 3-tuple, regardless of the dimension).
        """
        return self._store.size

    @property
    def format(self):
        """The texture format as a string. Usually a pygfx format specifier
        (e.g. u2 for scalar uint16, or 3xf4 for RGB float32),
        but can also be a overriden to a backend-specific format.
        """
        return self._format

    @property
    def colorspace(self):
        """If this data is used as color, it is interpreted to be in this colorspace.
        Can be "srgb" or "physical". Default "srgb".
        """
        return self._colorspace

    @property
    def generate_mipmaps(self):
        """Whether to automatically generate mipmaps when uploading to the GPU."""
        return self._generate_mipmaps

    def update_range(self, offset, size):
        """Mark a certain range of the data for upload to the GPU.
        The offset and (sub) size should be (width, height, depth)
        tuples. Numpy users beware that an arrays shape is (height, width)!
        """
        # Check input
        assert isinstance(offset, tuple) and len(offset) == 3
        assert isinstance(size, tuple) and len(size) == 3
        if any(s == 0 for s in size):
            return
        elif any(s < 0 for s in size):
            raise ValueError("Update size must not be negative")
        elif any(b < 0 for b in offset):
            raise ValueError("Update offset must not be negative")
        elif any(b + s > refsize for b, s, refsize in zip(offset, size, self.size)):
            raise ValueError("Update size out of range")
        # Apply - consider that texture arrays want to be uploaded per-texture
        # todo: avoid duplicates by merging with existing pending uploads
        if self.dim == 1:
            for z in range(size[2]):
                for y in range(size[1]):
                    offset2 = offset[0], y, z
                    size2 = size[0], 1, 1
                    self._gfx_pending_uploads.append((offset2, size2))
        elif self.dim == 2:
            for z in range(size[2]):
                offset2 = offset[0], offset[1], z
                size2 = size[0], size[1], 1
                self._gfx_pending_uploads.append((offset2, size2))
        else:
            self._gfx_pending_uploads.append((offset, size))
        Resource._rev += 1
        self._rev = Resource._rev
        self._gfx_mark_for_sync()

    def _size_from_data(self, data, dim, size):
        # Check if shape matches dimension
        shape = data.shape

        if size:
            # Get version of size with trailing ones stripped
            size2 = size
            size2 = size2[:-1] if size2[-1] == 1 else size2
            size2 = size2[:-1] if size2[-1] == 1 else size2
            rsize = tuple(reversed(size2))
            # Check if size matches shape
            if rsize != shape[: len(rsize)]:
                raise ValueError(f"Given size does not match the data shape.")
            return size
        else:
            if len(shape) not in (dim, dim + 1):
                raise ValueError(
                    f"Can't map shape {shape} on {dim}D tex. Maybe also specify size?"
                )
            # Determine size based on dim and shape
            if dim == 1:
                return shape[0], 1, 1
            elif dim == 2:
                return shape[1], shape[0], 1
            else:  # dim == 3:
                return shape[2], shape[1], shape[0]

    def _get_subdata(self, offset, size, pixel_padding=None):
        """Return subdata as a contiguous array."""
        # If this is a full range, this is easy
        if offset == 0 and size == self.nitems and self.mem.contiguous:
            return self.mem
        # Get a numpy array, because memoryviews do not support nd slicing
        if isinstance(self.data, np.ndarray):
            arr = self.data
        elif not self.mem.c_contiguous:
            raise ValueError(
                "Non-contiguous texture data is only supported for numpy array."
            )
        else:
            arr = np.frombuffer(self.mem, self.mem.format)
        arr = arr.reshape(self.size[2], self.size[1], self.size[0], -1)
        # Slice it
        slices = []
        for d in reversed(range(3)):
            slices.append(slice(offset[d], offset[d] + size[d]))
        sub_arr = arr[tuple(slices)]
        if pixel_padding is not None:
            padding = np.ones(sub_arr.shape[:3] + (1,), dtype=sub_arr.dtype)
            sub_arr = np.concatenate([sub_arr, pixel_padding * padding], -1)
        return memoryview(np.ascontiguousarray(sub_arr))

    def get_view(self, *args, **kwargs):
        raise DeprecationWarning(
            "Texture.get_view() is removed, TextureView is no longer public API: just use plain textures."
        )



import wgpu  # only for flags/enums

from . import register_wgpu_render_function, WorldObjectShader, Binding, RenderMask
from ._utils import to_texture_format, GfxSampler, GfxTextureView
from ...objects import Image
from ...materials import ImageBasicMaterial
from ...resources import Texture


vertex_and_fragment = wgpu.ShaderStage.VERTEX | wgpu.ShaderStage.FRAGMENT

sampled_value_to_color = """
        fn sampled_value_to_color(value_rgba: vec4<f32>) -> vec4<f32> {

            // Make it the correct dimension
            $$ if img_nchannels == 1
                let value_raw = value_rgba.r;
            $$ elif img_nchannels == 2
                let value_raw = value_rgba.rg;
            $$ elif img_nchannels == 3
                let value_raw = value_rgba.rgb;
            $$ else
                let value_raw = value_rgba.rgba;
            $$ endif

            // Apply contrast limits
            let value_cor = value_raw {{ climcorrection }};
            let value_clim = (value_cor - u_material.clim[0]) / (u_material.clim[1] - u_material.clim[0]);

            // Apply colormap or compose final color
            $$ if colormap_dim
                // In the render function we make sure that colormap_dim matches img_nchannels
                let color = sample_colormap(value_clim);
            $$ else
                $$ if img_nchannels == 1
                    let r = value_clim;
                    let color = vec4<f32>(r, r, r, 1.0);
                $$ elif img_nchannels == 2
                    let color = vec4<f32>(value_clim.rrr, value_raw.g);
                $$ elif img_nchannels == 3
                    let color = vec4<f32>(value_clim.rgb, 1.0);
                $$ else
                    let color = vec4<f32>(value_clim.rgb, value_raw.a);
                $$ endif
            $$ endif

            return color;
        }
"""


class BaseImageShader(WorldObjectShader):
    def code_image_helpers(self):
        return (
            sampled_value_to_color
            + """
        struct ImGeometry {
            indices: array<i32,6>,
            positions: array<vec3<f32>,4>,
            texcoords: array<vec2<f32>,4>,
        };

        fn get_im_geometry() -> ImGeometry {
            let size = textureDimensions(t_img);
            var geo: ImGeometry;

            geo.indices = array<i32,6>(0, 1, 2,   3, 2, 1);

            let pos1 = vec2<f32>(-0.5);
            let pos2 = vec2<f32>(size.xy) + pos1;
            geo.positions = array<vec3<f32>,4>(
                vec3<f32>(pos2.x, pos1.y, 0.0),
                vec3<f32>(pos2.x, pos2.y, 0.0),
                vec3<f32>(pos1.x, pos1.y, 0.0),
                vec3<f32>(pos1.x, pos2.y, 0.0),
            );

            geo.texcoords = array<vec2<f32>,4>(
                vec2<f32>(1.0, 0.0),
                vec2<f32>(1.0, 1.0),
                vec2<f32>(0.0, 0.0),
                vec2<f32>(0.0, 1.0),
            );

            return geo;
        }

        fn sample_im(texcoord: vec2<f32>, sizef: vec2<f32>) -> vec4<f32> {
            $$ if img_format == 'f32'
                return textureSample(t_img, s_img, texcoord.xy);
            $$ else
                let texcoords_u = vec2<i32>(texcoord.xy * sizef.xy);
                return vec4<f32>(textureLoad(t_img, texcoords_u, 0));
            $$ endif
        }
    """
        )


@register_wgpu_render_function(Image, ImageBasicMaterial)
class ImageShader(BaseImageShader):
    type = "render"

    def get_bindings(self, wobject, shared):
        geometry = wobject.geometry
        material = wobject.material  # noqa

        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
        ]

        self["climcorrection"] = ""

        # Collect texture and sampler
        if geometry.grid is None:
            raise ValueError("Image.geometry must have a grid (texture).")
        else:
            if not isinstance(geometry.grid, Texture):
                raise TypeError("Image.geometry.grid must be a Texture.")
            if geometry.grid.dim != 2:
                raise TypeError("Image.geometry.grid must a 2D texture")
            tex_view = GfxTextureView(geometry.grid)
            sampler = GfxSampler(material.interpolation, "clamp")
            self["colorspace"] = geometry.grid.colorspace
            # Sampling type
            fmt = to_texture_format(geometry.grid.format)
            if "norm" in fmt or "float" in fmt:
                self["img_format"] = "f32"
                if "unorm" in fmt:
                    self["climcorrection"] = " * 255.0"
                elif "snorm" in fmt:
                    self["climcorrection"] = " * 255.0 - 128.0"
            elif "uint" in fmt:
                self["img_format"] = "u32"
            else:
                self["img_format"] = "i32"
            # Channels
            self["img_nchannels"] = len(fmt) - len(fmt.lstrip("rgba"))

        bindings.append(Binding("s_img", "sampler/filtering", sampler, "FRAGMENT"))
        bindings.append(Binding("t_img", "texture/auto", tex_view, vertex_and_fragment))

        # If a colormap is applied ...
        if material.map is not None:
            bindings.extend(
                self.define_img_colormap(material.map, material.map_interpolation)
            )
            self["colorspace"] = material.map.colorspace

        bindings = {i: b for i, b in enumerate(bindings)}
        self.define_bindings(0, bindings)

        return {
            0: bindings,
        }

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_strip,
            "cull_mode": wgpu.CullMode.none,
        }

    def get_render_info(self, wobject, shared):
        material = wobject.material

        render_mask = wobject.render_mask
        if not render_mask:
            render_mask = RenderMask.all
            if material.is_transparent:
                render_mask = RenderMask.transparent
            elif material.map is not None:
                if self["colormap_nchannels"] in (1, 3):
                    render_mask = RenderMask.opaque
            else:
                if self["img_nchannels"] in (1, 3):
                    render_mask = RenderMask.opaque

        return {
            "indices": (4, 1),
            "render_mask": render_mask,
        }

    def get_code(self):
        return (
            self.code_definitions()
            + self.code_common()
            + self.code_image_helpers()
            + self.code_vertex()
            + self.code_fragment()
        )

    def code_vertex(self):
        return """

        struct VertexInput {
            @builtin(vertex_index) vertex_index : u32,
        };


        @vertex
        fn vs_main(in: VertexInput) -> Varyings {

            var geo = get_im_geometry();

            // Select what face we're at
            let index = i32(in.vertex_index);
            let i0 = geo.indices[index];

            // Sample position, and convert to world pos, and then to ndc
            let data_pos = vec4<f32>(geo.positions[i0], 1.0);
            let world_pos = u_wobject.world_transform * data_pos;
            let ndc_pos = u_stdinfo.projection_transform * u_stdinfo.cam_transform * world_pos;

            var varyings: Varyings;
            varyings.position = vec4<f32>(ndc_pos);
            varyings.world_pos = vec3<f32>(world_pos.xyz);
            varyings.texcoord = vec2<f32>(geo.texcoords[i0]);
            return varyings;
        }
        """

    def code_fragment(self):
        return """

        @fragment
        fn fs_main(varyings: Varyings) -> FragmentOutput {
            let sizef = vec2<f32>(textureDimensions(t_img));
            let value = sample_im(varyings.texcoord.xy, sizef);
            let color = sampled_value_to_color(value);

            // Move to physical colorspace (linear photon count) so we can do math
            $$ if colorspace == 'srgb'
                let physical_color = srgb2physical(color.rgb);
            $$ else
                let physical_color = color.rgb;
            $$ endif
            let opacity = color.a * u_material.opacity;
            let out_color = vec4<f32>(physical_color, opacity);

            // Wrap up
            apply_clipping_planes(varyings.world_pos);
            var out = get_fragment_output(varyings.position.z, out_color);

            $$ if write_pick
            // The wobject-id must be 20 bits. In total it must not exceed 64 bits.
            out.pick = (
                pick_pack(u32(u_wobject.id), 20) +
                pick_pack(u32(varyings.texcoord.x * 4194303.0), 22) +
                pick_pack(u32(varyings.texcoord.y * 4194303.0), 22)
            );
            $$ endif

            return out;
        }
        """



import wgpu  # only for flags/enums

from . import (
    register_wgpu_render_function,
    WorldObjectShader,
    Binding,
    RenderMask,
    shaderlib,
)
from ._utils import to_texture_format, GfxSampler, GfxTextureView
from ...objects import Mesh, InstancedMesh
from ...materials import (
    MeshBasicMaterial,
    MeshPhongMaterial,
    MeshNormalMaterial,
    MeshNormalLinesMaterial,
    MeshSliceMaterial,
    MeshStandardMaterial,
)
from ...resources import Buffer, Texture
from ...utils import normals_from_vertices


@register_wgpu_render_function(Mesh, MeshBasicMaterial)
class MeshShader(WorldObjectShader):
    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)

        material = wobject.material
        geometry = wobject.geometry

        # Is this an instanced mesh?
        self["instanced"] = isinstance(wobject, InstancedMesh)

        # Is this a wireframe mesh?
        self["wireframe"] = getattr(material, "wireframe", False)
        self["flat_shading"] = getattr(material, "flat_shading", False)

        # Lighting off in the base class
        self["lighting"] = ""
        self["receive_shadow"] = wobject.receive_shadow

        # Per-vertex color, colormap, or a plane color?
        self["colorspace"] = "srgb"

        color_mode = str(material.color_mode).split(".")[-1]
        if color_mode == "auto":
            if material.map is not None:
                self["color_mode"] = "vertex_map"
                self["color_buffer_channels"] = 0
                self["colorspace"] = material.map.colorspace
            else:
                self["color_mode"] = "uniform"
                self["color_buffer_channels"] = 0
        elif color_mode == "uniform":
            self["color_mode"] = "uniform"
            self["color_buffer_channels"] = 0
        elif color_mode == "vertex":
            self["color_mode"] = "vertex"
            self["color_buffer_channels"] = nchannels = geometry.colors.data.shape[1]
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        elif color_mode == "face":
            self["color_mode"] = "face"
            self["color_buffer_channels"] = nchannels = geometry.colors.data.shape[1]
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        elif color_mode == "vertex_map":
            self["color_mode"] = "vertex_map"
            self["color_buffer_channels"] = 0
            self["colorspace"] = material.map.colorspace
            if material.map is None:
                raise ValueError(f"Cannot apply colormap is no material.map is set.")
        elif color_mode == "face_map":
            self["color_mode"] = "face_map"
            self["color_buffer_channels"] = 0
            self["colorspace"] = material.map.colorspace
            if material.map is None:
                raise ValueError(f"Cannot apply colormap is no material.map is set.")
        else:
            raise RuntimeError(f"Unknown color_mode: '{color_mode}'")

    def get_bindings(self, wobject, shared):
        geometry = wobject.geometry
        material = wobject.material

        # We're assuming the presence of an index buffer for now
        assert getattr(geometry, "indices", None)

        # Triangles or quads?
        if geometry.indices.data is not None and geometry.indices.data.shape[-1] == 4:
            self["indexer"] = 6
        else:
            self["indexer"] = 3

        # Normals. Usually it'd be given. If not, we'll calculate it from the vertices.
        if getattr(geometry, "normals", None) is not None:
            normal_buffer = geometry.normals
        else:
            if geometry.indices.data.shape[-1] == 4:
                n1 = normals_from_vertices(
                    geometry.positions.data, geometry.indices.data[..., :3]
                )
                n2 = normals_from_vertices(
                    geometry.positions.data, geometry.indices.data[..., [0, 2, 3]]
                )
                normal_data = (n1 + n2) / 2
            else:
                normal_data = normals_from_vertices(
                    geometry.positions.data, geometry.indices.data
                )
            normal_buffer = Buffer(normal_data)

        # Init bindings
        rbuffer = "buffer/read_only_storage"
        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
            Binding("s_indices", rbuffer, geometry.indices, "VERTEX"),
            Binding("s_positions", rbuffer, geometry.positions, "VERTEX"),
            Binding("s_normals", rbuffer, normal_buffer, "VERTEX"),
        ]

        if hasattr(geometry, "texcoords1") and geometry.texcoords1 is not None:
            bindings.append(
                Binding(
                    "s_texcoords1",
                    "buffer/read_only_storage",
                    geometry.texcoords1,
                    "VERTEX",
                )
            )
            self["use_texcoords1"] = True

        if self["color_mode"] in ("vertex", "face"):
            bindings.append(Binding("s_colors", rbuffer, geometry.colors, "VERTEX"))
        elif self["color_mode"] in ("vertex_map", "face_map"):
            bindings.extend(
                self.define_texcoords_and_colormap(
                    material.map, geometry.texcoords, material.map_interpolation
                )
            )

        F = "FRAGMENT"  # noqa: N806
        sampling = "sampler/filtering"
        sampler = GfxSampler(material.map_interpolation, "repeat")
        texturing = "texture/auto"

        # set envmap configs
        if getattr(material, "env_map", None):
            self._check_texture(
                material.env_map, 6
            )  # TODO: envmap not only cube, but also equirect (hdr format)
            view = GfxTextureView(material.env_map, view_dim="cube")
            bindings.append(Binding("s_env_map", sampling, sampler, F))
            bindings.append(Binding("t_env_map", texturing, view, F))

            if isinstance(material, MeshStandardMaterial):
                self["use_IBL"] = True
            elif isinstance(material, MeshBasicMaterial):
                self["use_env_map"] = True
                self["env_combine_mode"] = getattr(
                    material, "env_combine_mode", "MULTIPLY"
                )

            self["env_mapping_mode"] = getattr(
                material, "env_mapping_mode", "CUBE-REFLECTION"
            )

        # set lightmap configs
        if getattr(material, "light_map", None):
            if "use_texcoords1" not in self.kwargs:
                raise ValueError(
                    "Light map requires a second set of texture coordinates (geometry.texcoords1), but it is not present."
                )
            else:
                self._check_texture(material.light_map)
                view = GfxTextureView(material.light_map, view_dim="2d")
                self["use_light_map"] = True
                bindings.append(Binding("s_light_map", sampling, sampler, F))
                bindings.append(Binding("t_light_map", texturing, view, F))

        # set aomap configs
        if getattr(material, "ao_map", None):
            if "use_texcoords1" not in self.kwargs:
                raise ValueError(
                    "AoMap requires a second set of texture coordinates (geometry.texcoords1), but it is not present."
                )
            else:
                self._check_texture(material.ao_map)
                view = GfxTextureView(material.ao_map, view_dim="2d")
                self["use_ao_map"] = True
                bindings.append(Binding("s_ao_map", sampling, sampler, F))
                bindings.append(Binding("t_ao_map", texturing, view, F))

        # Define shader code for binding
        bindings = {i: binding for i, binding in enumerate(bindings)}
        self.define_bindings(0, bindings)

        # Instanced meshes have an extra storage buffer that we add manually
        bindings1 = {}  # non-auto-generated bindings
        if self["instanced"]:
            bindings1[0] = Binding(
                "s_instance_infos", rbuffer, wobject.instance_buffer, "VERTEX"
            )

        return {
            0: bindings,
            1: bindings1,
        }

    def get_pipeline_info(self, wobject, shared):
        material = wobject.material

        topology = wgpu.PrimitiveTopology.triangle_list

        if material.side == "FRONT":
            cull_mode = wgpu.CullMode.back
        elif material.side == "BACK":
            cull_mode = wgpu.CullMode.front
        else:  # material.side == "BOTH"
            cull_mode = wgpu.CullMode.none

        return {
            "primitive_topology": topology,
            "cull_mode": cull_mode,
        }

    def get_render_info(self, wobject, shared):
        geometry = wobject.geometry
        material = wobject.material

        if geometry.indices.data is not None and geometry.indices.data.shape[-1] == 4:
            offset, size = geometry.indices.draw_range
            offset, size = 6 * offset, 6 * size
        else:
            offset, size = geometry.indices.draw_range
            offset, size = 3 * offset, 3 * size

        n_instances = 1
        if self["instanced"]:
            n_instances = wobject.instance_buffer.nitems

        render_mask = wobject.render_mask
        if not render_mask:
            render_mask = RenderMask.all
            if material.is_transparent:
                render_mask = RenderMask.transparent
            elif self["color_mode"] == "uniform":
                if material.color_is_transparent:
                    render_mask = RenderMask.transparent
                else:
                    render_mask = RenderMask.opaque
            elif self["color_mode"] in ("vertex", "face"):
                if self["color_buffer_channels"] in (1, 3):
                    render_mask = RenderMask.opaque
            elif self["color_mode"] in ("vertex_map", "face_map"):
                if self["colormap_nchannels"] in (1, 3):
                    render_mask = RenderMask.opaque
            elif self["color_mode"] == "normal":
                render_mask = RenderMask.opaque
            else:
                raise RuntimeError(f"Unexpected color mode {self['color_mode']}")

        return {
            "indices": (size, n_instances, offset, 0),
            "render_mask": render_mask,
        }

    def get_code(self):
        return (
            self.code_definitions()
            + self.code_common()
            + self.code_lighting()
            + self.code_vertex()
            + self.code_fragment()
        )

    def code_vertex(self):
        return """

        struct VertexInput {
            @builtin(vertex_index) vertex_index : u32,
            $$ if instanced
            @builtin(instance_index) instance_index : u32,
            $$ endif
        };

        $$ if instanced
        struct InstanceInfo {
            transform: mat4x4<f32>,
            id: u32,
        };
        @group(1) @binding(0)
        var<storage,read> s_instance_infos: array<InstanceInfo>;
        $$ endif

        fn get_sign_of_det_of_4x4(m: mat4x4<f32>) -> f32 {
            // We know/assume that the matrix is a homogeneous matrix,
            // so that only the 3x3 region is relevant for the determinant,
            // which is faster to calculate than the det of the 4x4.
            let m3 = mat3x3<f32>(m[0].xyz, m[1].xyz, m[2].xyz);
            return sign(determinant(m3));
        }

        fn dist_pt_line(x1: f32, y1: f32, x2: f32, y2: f32, x3: f32, y3: f32) -> f32 {
         // Distance of pt (x3,y3) to line with coords(x1,y1) (x2,y2)
         return abs((x2 - x1) * (y1 - y3) - (x1 - x3) * (y2 - y1)) / sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1));
        }

        @vertex
        fn vs_main(in: VertexInput) -> Varyings {

            // Get world transform
            $$ if instanced
                let instance_info = s_instance_infos[in.instance_index];
                let world_transform = u_wobject.world_transform * instance_info.transform;
            $$ else
                let world_transform = u_wobject.world_transform;
            $$ endif

            // Select what face we're at
            let index = i32(in.vertex_index);
            let face_index = index / {{indexer}};
            var sub_index = index % {{indexer}};
            var face_sub_index = 0;

            // for quads assuming the vertices are oriented, the triangles are 0 1 2 and 0 2 3
            $$ if indexer == 6
                var quad_map = array<i32,6>(0, 1, 2, 0, 2, 3);
                //face_sub_index returns 0 or 1 in picking for quads. So the triangle of the quad can be identified.
                var face_map = array<i32,6>(0, 0, 0, 1, 1, 1);
                face_sub_index = face_map[sub_index];
                sub_index = quad_map[sub_index];
            $$ endif

            // If a transform has an uneven number of negative scales, the 3 vertices
            // that make up the face are such that the GPU will mix up front and back
            // faces, producing an incorrect is_front. We can detect this from the
            // sign of the determinant, and reorder the faces to fix it. Note that
            // the projection_transform is not included here, because it cannot be
            // set with the public API and we assume that it does not include a flip.
            let winding_world = get_sign_of_det_of_4x4(world_transform);
            let winding_cam = get_sign_of_det_of_4x4(u_stdinfo.cam_transform);
            let must_flip_sub_index = winding_world * winding_cam < 0.0;
            // If necessary, and the sub_index is even, e.g. 0 or 2, we flip it to the other.
            // Flipping 0 and 2, because they are present in both triangles of a quad.
            if (must_flip_sub_index && sub_index % 2 == 0) {
                sub_index = select(0, 2, sub_index == 0);
            }

            // Sample
            let vii = load_s_indices(face_index);
            let i0 = i32(vii[sub_index]);

            // Get vertex position
            let raw_pos = load_s_positions(i0);
            let world_pos = world_transform * vec4<f32>(raw_pos, 1.0);
            var ndc_pos = u_stdinfo.projection_transform * u_stdinfo.cam_transform * world_pos;

            // For the wireframe we also need the ndc_pos of the other vertices of this face
            $$ if wireframe
                $$ for i in ((1, 2, 3) if indexer == 3 else (1, 2, 3, 4))
                    let raw_pos{{ i }} = load_s_positions(i32(vii[{{ i - 1 }}]));
                    let world_pos{{ i }} = world_transform * vec4<f32>(raw_pos{{ i }}, 1.0);
                    let ndc_pos{{ i }} = u_stdinfo.projection_transform * u_stdinfo.cam_transform * world_pos{{ i }};
                $$ endfor
                let depth_offset = -0.0001;  // to put the mesh slice atop a mesh
                ndc_pos.z = ndc_pos.z + depth_offset;
            $$ endif

            // Prepare output
            var varyings: Varyings;

            // Set position
            varyings.world_pos = vec3<f32>(world_pos.xyz / world_pos.w);
            varyings.position = vec4<f32>(ndc_pos.xyz, ndc_pos.w);

            // per-vertex or per-face coloring
            $$ if color_mode == 'face' or color_mode == 'vertex'
                $$ if color_mode == 'face'
                    let color_index = face_index;
                $$ else
                    let color_index = i0;
                $$ endif
                $$ if color_buffer_channels == 1
                    let cvalue = load_s_colors(color_index);
                    varyings.color = vec4<f32>(cvalue, cvalue, cvalue, 1.0);
                $$ elif color_buffer_channels == 2
                    let cvalue = load_s_colors(color_index);
                    varyings.color = vec4<f32>(cvalue.r, cvalue.r, cvalue.r, cvalue.g);
                $$ elif color_buffer_channels == 3
                    varyings.color = vec4<f32>(load_s_colors(color_index), 1.0);
                $$ elif color_buffer_channels == 4
                    varyings.color = vec4<f32>(load_s_colors(color_index));
                $$ endif
            $$ endif

            // How to index into tex-coords
            $$ if color_mode == 'face_map'
            let tex_coord_index = face_index;
            $$ else
            let tex_coord_index = i0;
            $$ endif

            // Set texture coords
            $$ if colormap_dim == '1d'
            varyings.texcoord = f32(load_s_texcoords(tex_coord_index));
            $$ elif colormap_dim == '2d'
            varyings.texcoord = vec2<f32>(load_s_texcoords(tex_coord_index));
            $$ elif colormap_dim == '3d'
            varyings.texcoord = vec3<f32>(load_s_texcoords(tex_coord_index));
            $$ endif

            $$ if use_texcoords1 is defined
            varyings.texcoord1 = vec2<f32>(load_s_texcoords1(tex_coord_index));
            $$ endif

            // Set the normal
            let raw_normal = load_s_normals(i0);
            // Transform the normal to world space
            // Note that the world transform matrix cannot be directly applied to the normal
            let normal_matrix = transpose(u_wobject.world_transform_inv);
            let world_normal = normalize((normal_matrix * vec4<f32>(raw_normal, 0.0)).xyz);

            varyings.normal = vec3<f32>(world_normal);
            varyings.geometry_normal = vec3<f32>(raw_normal);
            varyings.winding_cam = f32(winding_cam);

            // Set wireframe barycentric-like coordinates
            $$ if wireframe
                $$ if indexer == 3
                    $$ for i in (1, 2, 3)
                        let p{{ i }} = (ndc_pos{{ i }}.xy / ndc_pos{{ i }}.w) * u_stdinfo.logical_size * 0.5;
                    $$ endfor
                    let dist1 = dist_pt_line(p2.x,p2.y,p3.x,p3.y,p1.x,p1.y);
                    let dist2 = dist_pt_line(p1.x,p1.y,p3.x,p3.y,p2.x,p2.y);
                    let dist3 = dist_pt_line(p1.x,p1.y,p2.x,p2.y,p3.x,p3.y);
                    var arr_wireframe_coords = array<vec3<f32>, 3>(
                        vec3<f32>(dist1, 0.0, 0.0), vec3<f32>(0.0, dist2, 0.0), vec3<f32>(0.0, 0.0, dist3)
                    );
                    varyings.wireframe_coords = vec3<f32>(arr_wireframe_coords[sub_index]);  // in logical pixels
                $$ elif indexer == 6
                    $$ for i in (1, 2, 3, 4)
                        let p{{ i }} = (ndc_pos{{ i }}.xy / ndc_pos{{ i }}.w) * u_stdinfo.logical_size * 0.5;
                    $$ endfor
                    //dist of vertex 1 to segment 23
                    let dist1_23 = dist_pt_line(p2.x,p2.y,p3.x,p3.y,p1.x,p1.y);
                    //dist of vertex 1 to segment 34
                    let dist1_34 = dist_pt_line(p3.x,p3.y,p4.x,p4.y,p1.x,p1.y);

                    //dist of vertex 2 to segment 34
                    let dist2_34 = dist_pt_line(p3.x,p3.y,p4.x,p4.y,p2.x,p2.y);
                    //dist of vertex 2 to segment 14
                    let dist2_14 = dist_pt_line(p1.x,p1.y,p4.x,p4.y,p2.x,p2.y);

                    //dist of vertex 3 to segment 12
                    let dist3_12 = dist_pt_line(p1.x,p1.y,p2.x,p2.y,p3.x,p3.y);
                    //dist of vertex 3 to segment 14
                    let dist3_14 = dist_pt_line(p1.x,p1.y,p4.x,p4.y,p3.x,p3.y);

                    //dist of vertex 4 to segment 12
                    let dist4_12 = dist_pt_line(p2.x,p2.y,p1.x,p1.y,p4.x,p4.y);
                    //dist of vertex 4 to segment 23
                    let dist4_23 = dist_pt_line(p2.x,p2.y,p3.x,p3.y,p4.x,p4.y);

                    //segments 12 23 34 41
                    var arr_wireframe_coords = array<vec4<f32>, 4>(
                        vec4<f32>( 0.0, dist1_23,dist1_34, 0.0),
                        vec4<f32>(0.0, 0.0, dist2_34, dist2_14),
                        vec4<f32>( dist3_12 ,0.0, 0.0, dist3_14),
                        vec4<f32>( dist4_12,dist4_23, 0.0, 0.0)
                        );
                    varyings.wireframe_coords = vec4<f32>(arr_wireframe_coords[sub_index]);  // in logical pixels
                $$ endif
            $$ endif

            // Set varyings for picking. We store the face_index, and 3 weights
            // that indicate how close the fragment is to each vertex (barycentric
            // coordinates). This allows the selection of the nearest vertex or edge.
            $$ if instanced
                let pick_id = instance_info.id;
            $$ else
                let pick_id = u_wobject.id;
            $$ endif

            varyings.pick_id = u32(pick_id);
            $$ if indexer == 3
            varyings.pick_idx = u32(face_index);
            $$ else
            varyings.pick_idx = u32(face_index * 2 + face_sub_index);
            $$ endif

            var arr_pick_coords = array<vec3<f32>, 4>(vec3<f32>(1.0, 0.0, 0.0),
                                                      vec3<f32>(0.0, 1.0, 0.0),
                                                      vec3<f32>(0.0, 0.0, 1.0),
                                                      vec3<f32>(0.0, 1.0, 0.0),  // the 2nd triangle in a quad
                                                      );
            varyings.pick_coords = vec3<f32>(arr_pick_coords[sub_index]);

            return varyings;
        }

    """

    def code_fragment(self):
        return """

        @fragment
        fn fs_main(varyings: Varyings, @builtin(front_facing) is_front: bool) -> FragmentOutput {

            // Get the surface normal from the geometry.
            // This is the unflipped normal, because thet NormalMaterial needs that.
            var surface_normal = normalize(vec3<f32>(varyings.normal));
            $$ if flat_shading
                let u = dpdx(varyings.world_pos);
                let v = dpdy(varyings.world_pos);
                surface_normal = normalize(cross(u, v));
                // Because this normal is derived from the world_pos, it has been corrected
                // for some of the winding, but not all. We apply the below steps to
                // bring it in the same state as the regular (non-flat) shading.
                surface_normal = select(-surface_normal, surface_normal, varyings.winding_cam < 0.0);
                surface_normal = select(-surface_normal, surface_normal, is_front);
            $$ endif

            $$ if color_mode == 'vertex' or color_mode == 'face'
                let color_value = varyings.color;
                let albeido = color_value.rgb;
            $$ elif color_mode == 'vertex_map' or color_mode == 'face_map'
                let color_value = sample_colormap(varyings.texcoord);
                let albeido = color_value.rgb;  // no more colormap
            $$ elif color_mode == 'normal'
                let albeido = normalize(surface_normal) * 0.5 + 0.5;
                let color_value = vec4<f32>(albeido, 1.0);
            $$ else
                let color_value = u_material.color;
                let albeido = color_value.rgb;
            $$ endif

            // Move to physical colorspace (linear photon count) so we can do math
            $$ if colorspace == 'srgb'
                let physical_albeido = srgb2physical(albeido);
            $$ else
                let physical_albeido = albeido;
            $$ endif
            let opacity = color_value.a * u_material.opacity;

            // Get normal used to calculate lighting or reflection
            $$ if lighting or use_env_map is defined
                // Get view direction
                let view = select(
                    normalize(u_stdinfo.cam_transform_inv[3].xyz - varyings.world_pos),
                    ( u_stdinfo.cam_transform_inv * vec4<f32>(0.0, 0.0, 1.0, 0.0) ).xyz,
                    is_orthographic()
                );
                // Get normal used to calculate lighting
                var normal = select(-surface_normal, surface_normal, is_front);
                $$ if use_normal_map is defined
                    let normal_map = textureSample( t_normal_map, s_normal_map, varyings.texcoord ) * 2.0 - 1.0;
                    let normal_map_scale = vec3<f32>( normal_map.xy * u_material.normal_scale, normal_map.z );
                    normal = perturbNormal2Arb(view, normal, normal_map_scale, varyings.texcoord, is_front);
                $$ endif
            $$ endif

            // Lighting
            $$ if lighting
                // Do the math
                var physical_color = lighting_{{ lighting }}(varyings, normal, view, physical_albeido);
            $$ else
                var physical_color = physical_albeido;

                // Light map (pre-baked lighting)
                $$ if use_light_map is defined
                    let light_map_color = srgb2physical( textureSample( t_light_map, s_light_map, varyings.texcoord1 ).rgb );
                    physical_color *= light_map_color * u_material.light_map_intensity * RECIPROCAL_PI;
                $$ endif

                // Ambient occlusion
                $$ if use_ao_map is defined
                    let ao_map_intensity = u_material.ao_map_intensity;
                    let ambientOcclusion = ( textureSample( t_ao_map, s_ao_map, varyings.texcoord1 ).r - 1.0 ) * ao_map_intensity + 1.0;
                    physical_color *= ambientOcclusion;
                $$ endif
            $$ endif

            // Environment mapping
            $$ if use_env_map is defined
                let reflectivity = u_material.reflectivity;
                let specular_strength = 1.0; // TODO: support specular_map
                $$ if env_mapping_mode == "CUBE-REFLECTION"
                    var reflectVec = reflect( -view, normal );
                $$ elif env_mapping_mode == "CUBE-REFRACTION"
                    var reflectVec = refract( -view, normal, u_material.refraction_ratio );
                $$ endif
                var env_color_srgb = textureSample( t_env_map, s_env_map, vec3<f32>( -reflectVec.x, reflectVec.yz) );
                let env_color = srgb2physical(env_color_srgb.rgb); // TODO: maybe already in linear-space
                $$ if env_combine_mode == 'MULTIPLY'
                    physical_color = mix(physical_color, physical_color * env_color.xyz, specular_strength * reflectivity);
                $$ elif env_combine_mode == 'MIX'
                    physical_color = mix(physical_color, env_color.xyz, specular_strength * reflectivity);
                $$ elif env_combine_mode == 'ADD'
                    physical_color = physical_color + env_color.xyz * specular_strength * reflectivity;
                $$ endif
            $$ endif

            $$ if wireframe
                $$ if indexer == 3
                let distance_from_edge = min(varyings.wireframe_coords.x, min(varyings.wireframe_coords.y, varyings.wireframe_coords.z));
                $$ else
                let distance_from_edge = min(varyings.wireframe_coords.x, min(varyings.wireframe_coords.y, min(varyings.wireframe_coords.z, varyings.wireframe_coords.a)));
                $$ endif
                if (distance_from_edge > 0.5 * u_material.wireframe) {
                    discard;
                }
            $$ endif

            let out_color = vec4<f32>(physical_color, opacity);

            // Wrap up

            apply_clipping_planes(varyings.world_pos);
            var out = get_fragment_output(varyings.position.z, out_color);

            $$ if write_pick
            // The wobject-id must be 20 bits. In total it must not exceed 64 bits.
            out.pick = (
                pick_pack(varyings.pick_id, 20) +
                pick_pack(varyings.pick_idx, 26) +
                pick_pack(u32(varyings.pick_coords.x * 63.0), 6) +
                pick_pack(u32(varyings.pick_coords.y * 63.0), 6) +
                pick_pack(u32(varyings.pick_coords.z * 63.0), 6)
            );
            $$ endif

            return out;
        }

        """

    def code_lighting(self):
        return ""

    def _check_texture(self, t, size2=1):
        assert isinstance(t, Texture)
        assert t.size[2] == size2
        fmt = to_texture_format(t.format)
        assert "norm" in fmt or "float" in fmt


@register_wgpu_render_function(Mesh, MeshNormalMaterial)
class MeshNormalShader(MeshShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        self["color_mode"] = "normal"
        self["colormap_dim"] = ""  # disable texture if there happens to be one


@register_wgpu_render_function(Mesh, MeshPhongMaterial)
class MeshPhongShader(MeshShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        self["lighting"] = "phong"

    def code_lighting(self):
        # return shaderlib.lighting_phong_simple()  # the quick 'n dirty way
        code = ""
        if self["receive_shadow"]:
            code += shaderlib.shadow()
        code += shaderlib.lighting_phong()
        return code


@register_wgpu_render_function(Mesh, MeshStandardMaterial)
class MeshStandardShader(MeshShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        self["lighting"] = "pbr"

    def get_bindings(self, wobject, shared):
        result = super().get_bindings(wobject, shared)

        geometry = wobject.geometry
        material = wobject.material

        bindings = []

        F = "FRAGMENT"  # noqa: N806
        sampling = "sampler/filtering"
        sampler = GfxSampler(material.map_interpolation, "repeat")
        texturing = "texture/auto"

        # We need uv to use the maps, so if uv not exist, ignore all maps
        if hasattr(geometry, "texcoords") and geometry.texcoords is not None:
            # Texcoords must always be nx2 since it used for all texture maps.
            if not (
                geometry.texcoords.data.ndim == 2
                and geometry.texcoords.data.shape[1] == 2
            ):
                raise ValueError("For standard material, the texcoords must be Nx2")

            if material.normal_map is not None:
                self._check_texture(material.normal_map)
                view = GfxTextureView(material.normal_map, view_dim="2d")
                self["use_normal_map"] = True
                bindings.append(Binding("s_normal_map", sampling, sampler, F))
                bindings.append(Binding("t_normal_map", texturing, view, F))

            if material.roughness_map is not None:
                self._check_texture(material.roughness_map)
                view = GfxTextureView(material.roughness_map, view_dim="2d")
                self["use_roughness_map"] = True
                bindings.append(Binding("s_roughness_map", sampling, sampler, F))
                bindings.append(Binding("t_roughness_map", texturing, view, F))

            if material.metalness_map is not None:
                self._check_texture(material.metalness_map)
                view = GfxTextureView(material.metalness_map, view_dim="2d")
                self["use_metalness_map"] = True
                bindings.append(Binding("s_metalness_map", sampling, sampler, F))
                bindings.append(Binding("t_metalness_map", texturing, view, F))

            if material.emissive_map is not None:
                self._check_texture(material.emissive_map)
                view = GfxTextureView(material.emissive_map, view_dim="2d")
                self["use_emissive_map"] = True
                bindings.append(Binding("s_emissive_map", sampling, sampler, F))
                bindings.append(Binding("t_emissive_map", texturing, view, F))

        # Define shader code for binding
        bindings = {i: binding for i, binding in enumerate(bindings)}
        self.define_bindings(2, bindings)

        # Update result
        result[2] = bindings
        return result

    def code_lighting(self):
        code = ""
        if self["receive_shadow"]:
            code += shaderlib.shadow()
        code += shaderlib.lighting_pbr()
        return code


@register_wgpu_render_function(Mesh, MeshNormalLinesMaterial)
class MeshNormalLinesShader(MeshShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        self["color_mode"] = "uniform"
        self["lighting"] = ""
        self["wireframe"] = False

    def get_pipeline_info(self, wobject, shared):
        d = super().get_pipeline_info(wobject, shared)
        d["primitive_topology"] = wgpu.PrimitiveTopology.line_list
        return d

    def get_render_info(self, wobject, shared):
        # We directly look at the vertex data, so geometry.indices.draw_range is ignored.
        d = super().get_render_info(wobject, shared)
        d["indices"] = wobject.geometry.positions.nitems * 2, d["indices"][1]
        return d

    def code_vertex(self):
        return """

        struct VertexInput {
            @builtin(vertex_index) vertex_index : u32,
        };


        @vertex
        fn vs_main(in: VertexInput) -> Varyings {
            let index = i32(in.vertex_index);
            let r = index % 2;
            let i0 = index / 2;

            // Get regular position
            let raw_pos = load_s_positions(i0);
            var world_pos = u_wobject.world_transform * vec4<f32>(raw_pos, 1.0);

            // Get the normal, expressed in world coords. Use the normal-matrix
            // to take anisotropic scaling into account.
            let normal_matrix = transpose(u_wobject.world_transform_inv);
            let raw_normal = load_s_normals(i0);
            let world_normal = normalize((normal_matrix * vec4<f32>(raw_normal, 0.0)).xyz);

            // Calculate the two end-pieces of the line that we want to show.
            let pos1 = world_pos.xyz / world_pos.w;
            let pos2 = pos1 + world_normal * u_material.line_length;

            // Select either end of the line and make this the world pos
            let pos3 = pos1 * f32(r) + pos2 * (1.0 - f32(r));
            world_pos = vec4<f32>(pos3 * world_pos.w, world_pos.w,);

            // To NDC
            let ndc_pos = u_stdinfo.projection_transform * u_stdinfo.cam_transform * world_pos;

            var varyings: Varyings;
            varyings.world_pos = vec3<f32>(world_pos.xyz / world_pos.w);
            varyings.position = vec4<f32>(ndc_pos);

            // Stub varyings, because the mesh varyings are based on face index
            varyings.normal = vec3<f32>(world_normal);
            varyings.pick_id = u32(u_wobject.id);
            varyings.pick_idx = u32(0);
            varyings.pick_coords = vec3<f32>(0.0);

            return varyings;
        }
        """


@register_wgpu_render_function(Mesh, MeshSliceMaterial)
class MeshSliceShader(WorldObjectShader):
    """Shader for rendering mesh slices."""

    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)

        material = wobject.material
        geometry = wobject.geometry

        color_mode = str(material.color_mode).split(".")[-1]
        if color_mode == "auto":
            if material.map is not None:
                self["color_mode"] = "vertex_map"
                self["color_buffer_channels"] = 0
            else:
                self["color_mode"] = "uniform"
                self["color_buffer_channels"] = 0
        elif color_mode == "uniform":
            self["color_mode"] = "uniform"
            self["color_buffer_channels"] = 0
        elif color_mode == "vertex":
            self["color_mode"] = "vertex"
            self["color_buffer_channels"] = nchannels = geometry.colors.data.shape[1]
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        elif color_mode == "face":
            self["color_mode"] = "face"
            self["color_buffer_channels"] = nchannels = geometry.colors.data.shape[1]
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        elif color_mode == "vertex_map":
            self["color_mode"] = "vertex_map"
            self["color_buffer_channels"] = 0
            if material.map is None:
                raise ValueError(f"Cannot apply colormap is no material.map is set.")
        elif color_mode == "face_map":
            self["color_mode"] = "face_map"
            self["color_buffer_channels"] = 0
            if material.map is None:
                raise ValueError(f"Cannot apply colormap is no material.map is set.")
        else:
            raise RuntimeError(f"Unknown color_mode: '{color_mode}'")

    def get_bindings(self, wobject, shared):
        # It would technically be possible to implement colormapping or
        # per-vertex colors, but its a tricky dance to get the per-vertex
        # data (e.g. texcoords) into a varying. And because the visual
        # result is a line, its likely that in most use-cases a uniform
        # color is preferred anyway. So for now we don't implement that.

        geometry = wobject.geometry
        material = wobject.material

        # Init uniform bindings
        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
        ]

        # We're assuming the presence of an index buffer for now
        assert getattr(geometry, "indices", None)

        # Init storage buffer bindings
        rbuffer = "buffer/read_only_storage"
        bindings.append(Binding("s_indices", rbuffer, geometry.indices, "VERTEX"))
        bindings.append(Binding("s_positions", rbuffer, geometry.positions, "VERTEX"))

        # Bindings for color
        if self["color_mode"] in ("vertex", "face"):
            bindings.append(Binding("s_colors", rbuffer, geometry.colors, "VERTEX"))
        elif self["color_mode"] in ("vertex_map", "face_map"):
            bindings.extend(
                self.define_texcoords_and_colormap(
                    material.map, geometry.texcoords, material.map_interpolation
                )
            )

        # Let the shader generate code for our bindings
        bindings = {i: binding for i, binding in enumerate(bindings)}
        self.define_bindings(0, bindings)

        return {
            0: bindings,
        }

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_list,
            "cull_mode": wgpu.CullMode.none,
        }

    def get_render_info(self, wobject, shared):
        material = wobject.material  # noqa
        geometry = wobject.geometry

        offset, size = geometry.indices.draw_range
        offset, size = offset * 6, size * 6

        # As long as we don't use alpha for aa in the frag shader, we can use a render_mask of 1 or 2.
        render_mask = wobject.render_mask
        if not render_mask:
            if material.is_transparent:
                render_mask = RenderMask.transparent
            elif self["color_mode"] == "uniform":
                if material.color_is_transparent:
                    render_mask = RenderMask.transparent
                else:
                    render_mask = RenderMask.opaque
            elif self["color_mode"] in ("vertex", "face"):
                if self["color_buffer_channels"] in (1, 3):
                    render_mask = RenderMask.opaque
                else:
                    render_mask = RenderMask.all
            elif self["color_mode"] in ("vertex_map", "face_map"):
                if self["colormap_nchannels"] in (1, 3):
                    render_mask = RenderMask.opaque
                else:
                    render_mask = RenderMask.all
            else:
                raise RuntimeError(f"Unexpected color mode {self['color_mode']}")

        return {
            "indices": (size, 1, offset, 0),
            "render_mask": render_mask,
        }

    def get_code(self):
        return (
            self.code_definitions()
            + self.code_common()
            + self.code_vertex()
            + self.code_fragment()
        )

    def code_vertex(self):
        return """
        struct VertexInput {
            @builtin(vertex_index) vertex_index : u32,
        };
        @vertex
        fn vs_main(in: VertexInput) -> Varyings {
            // This vertex shader uses VertexId and storage buffers instead of
            // vertex buffers. It creates 6 vertices for each face in the mesh,
            // drawn with triangle-list. For the faces that cross the plane, we
            // draw a (thick) line segment with round caps (we need 6 verts for that).
            // Other faces become degenerate triangles.
            let screen_factor = u_stdinfo.logical_size.xy / 2.0;
            let l2p = u_stdinfo.physical_size.x / u_stdinfo.logical_size.x;
            let thickness = u_material.thickness;  // in logical pixels

            // Get the face index, and sample the vertex indices
            let index = i32(in.vertex_index);
            let segment_index = index % 6;
            let face_index = (index - segment_index) / 6;
            let vii = vec3<i32>(load_s_indices(face_index));

            // Vertex positions of this face, in local object coordinates
            let pos1a = load_s_positions(vii[0]);
            let pos2a = load_s_positions(vii[1]);
            let pos3a = load_s_positions(vii[2]);
            let pos1b = u_wobject.world_transform * vec4<f32>(pos1a, 1.0);
            let pos2b = u_wobject.world_transform * vec4<f32>(pos2a, 1.0);
            let pos3b = u_wobject.world_transform * vec4<f32>(pos3a, 1.0);
            let pos1 = pos1b.xyz / pos1b.w;
            let pos2 = pos2b.xyz / pos2b.w;
            let pos3 = pos3b.xyz / pos3b.w;
            // Get the plane definition
            let plane = u_material.plane.xyzw;  // ax + by + cz + d == 0
            let n = plane.xyz;  // not necessarily a unit vector
            // Intersect the plane with pos 1 and 2
            var p: vec3<f32>;
            p = pos1.xyz;
            let denom1 = dot(n,  pos2.xyz - pos1.xyz);
            let t1 = -(plane.x * p.x + plane.y * p.y + plane.z * p.z + plane.w) / denom1;
            // Intersect the plane with pos 2 and 3
            p = pos2.xyz;
            let denom2 = dot(n, pos3.xyz - pos2.xyz);
            let t2 = -(plane.x * p.x + plane.y * p.y + plane.z * p.z + plane.w) / denom2;
            // Intersect the plane with pos 3 and 1
            p = pos3.xyz;
            let denom3 = dot(n, pos1.xyz - pos3.xyz);
            let t3 = -(plane.x * p.x + plane.y * p.y + plane.z * p.z + plane.w) / denom3;
            // Selectors (the denom check seems not needed, but it feels safer to do, in case e.g. the precision changes)
            let b1 = select(0, 4, (t1 >= 0.0) && (t1 <= 1.0) && (denom1 != 0.0));
            let b2 = select(0, 2, (t2 >= 0.0) && (t2 <= 1.0) && (denom2 != 0.0));
            let b3 = select(0, 1, (t3 >= 0.0) && (t3 <= 1.0) && (denom3 != 0.0));
            var pos_index: i32;
            pos_index = b1 + b2 + b3;

            // The big triage
            var the_pos: vec4<f32>;
            var the_coord: vec2<f32>;
            var segment_length: f32;
            var pick_idx = u32(0u);
            var pick_coords = vec3<f32>(0.0);
            if (pos_index < 3 || pos_index == 4) {
                // Just return the same vertex, resulting in degenerate triangles
                the_pos = u_stdinfo.projection_transform * u_stdinfo.cam_transform * vec4<f32>(pos1, 1.0);
                the_coord = vec2<f32>(0.0, 0.0);
                segment_length = 0.0;
            } else {
                if (pos_index == 7) {
                    // For each edge we check whether it looks like it is selected. The value
                    // of the other t does not matter because it's always on the edge too.
                    // I would expect that comparing with 0.0 and 1.0 would work, but
                    // apparently the t-values can also be 0.5.
                    if (t1 < 0.5 && t2 >= 0.5) { pos_index = 6; }
                    else if (t2 < 0.5 && t3 >= 0.5) { pos_index = 3; }
                    else if (t3 < 0.5 && t1 >= 0.5) { pos_index = 5; }
                }
                // Get the positions where the frame intersects the plane
                let pos00: vec3<f32> = pos1;
                let pos12: vec3<f32> = mix(pos1, pos2, vec3<f32>(t1, t1, t1));
                let pos23: vec3<f32> = mix(pos2, pos3, vec3<f32>(t2, t2, t2));
                let pos31: vec3<f32> = mix(pos3, pos1, vec3<f32>(t3, t3, t3));
                // b1+b2+b3     000    001    010    011    100    101    110    111
                var positions_a = array<vec3<f32>, 8>(pos00, pos00, pos00, pos23, pos00, pos12, pos12, pos00);
                var positions_b = array<vec3<f32>, 8>(pos00, pos00, pos00, pos31, pos00, pos31, pos23, pos00);
                // Select the two positions that define the line segment
                let pos_a = positions_a[pos_index];
                let pos_b = positions_b[pos_index];
                // Same for face weights
                let fw00 = vec3<f32>(0.5, 0.5, 0.5);
                let fw12 = mix(vec3<f32>(1.0, 0.0, 0.0), vec3<f32>(0.0, 1.0, 0.0), vec3<f32>(t1, t1, t1));
                let fw23 = mix(vec3<f32>(0.0, 1.0, 0.0), vec3<f32>(0.0, 0.0, 1.0), vec3<f32>(t2, t2, t2));
                let fw31 = mix(vec3<f32>(0.0, 0.0, 1.0), vec3<f32>(1.0, 0.0, 0.0), vec3<f32>(t3, t3, t3));
                var fws_a = array<vec3<f32>, 8>(fw00, fw00, fw00, fw23, fw00, fw12, fw12, fw00);
                var fws_b = array<vec3<f32>, 8>(fw00, fw00, fw00, fw31, fw00, fw31, fw23, fw00);
                let fw_a = fws_a[pos_index];
                let fw_b = fws_b[pos_index];
                // Go from local coordinates to NDC
                var npos_a: vec4<f32> = u_stdinfo.projection_transform * u_stdinfo.cam_transform * vec4<f32>(pos_a, 1.0);
                var npos_b: vec4<f32> = u_stdinfo.projection_transform * u_stdinfo.cam_transform * vec4<f32>(pos_b, 1.0);
                // Don't forget to "normalize"!
                // todo: omitting this step diminish the thickness with distance, but it that the way?
                npos_a = npos_a / npos_a.w;
                npos_b = npos_b / npos_b.w;
                // And to logical pixel coordinates (don't worry about offset)
                let ppos_a = npos_a.xy * screen_factor;
                let ppos_b = npos_b.xy * screen_factor;
                // Get the segment vector, its length, and how much it scales because of thickness
                let v0 = ppos_b - ppos_a;
                segment_length = length(v0);  // in logical pixels;
                let segment_factor = (segment_length + thickness) / segment_length;
                // Get the (orthogonal) unit vectors that span the segment
                let v1 = normalize(v0);
                let v2 = vec2<f32>(v1.y, -v1.x);
                // Get the vector, in local logical pixels for the segment's square
                let pvec_local = 0.5 * vec2<f32>(segment_length + thickness, thickness);
                // Select one of the four corners of the segment rectangle
                var vecs = array<vec2<f32>, 6>(
                    vec2<f32>(-1.0, -1.0),
                    vec2<f32>( 1.0,  1.0),
                    vec2<f32>(-1.0,  1.0),
                    vec2<f32>( 1.0,  1.0),
                    vec2<f32>(-1.0, -1.0),
                    vec2<f32>( 1.0, -1.0),
                );
                let the_vec = vecs[segment_index];
                // Construct the position, also make sure zw scales correctly
                let pvec = the_vec.x * pvec_local.x * v1 + the_vec.y * pvec_local.y * v2;
                let z_range = (npos_b.z - npos_a.z) * segment_factor;
                let the_pos_p = 0.5 * (ppos_a + ppos_b) + pvec;
                let the_pos_z = 0.5 * (npos_a.z + npos_b.z) + the_vec.x * z_range * 0.5;
                let depth_offset = -0.0001;  // to put the mesh slice atop a mesh
                the_pos = vec4<f32>(the_pos_p / screen_factor, the_pos_z + depth_offset, 1.0);
                // Define the local coordinate in physical pixels
                the_coord = the_vec * pvec_local;
                // Picking info
                pick_idx = u32(face_index);
                let mixval = the_vec.x * 0.5 + 0.5;
                pick_coords = vec3<f32>(mix(fw_a, fw_b, vec3<f32>(mixval, mixval, mixval)));
            }

            // Shader output
            var varyings: Varyings;
            varyings.position = vec4<f32>(the_pos);
            varyings.world_pos = vec3<f32>(ndc_to_world_pos(the_pos));
            varyings.dist2center = vec2<f32>(the_coord * l2p);
            varyings.segment_length = f32(segment_length * l2p);
            varyings.segment_width = f32(thickness * l2p);
            varyings.pick_idx = u32(pick_idx);
            varyings.pick_coords = vec3<f32>(pick_coords);

            // per-vertex or per-face coloring
            $$ if color_mode == 'face' or color_mode == 'vertex'
                $$ if color_mode == 'face'
                    let cvalue = load_s_colors(face_index);
                $$ else
                    let cvalue = (  load_s_colors(vii[0]) * pick_coords[0] +
                                    load_s_colors(vii[1]) * pick_coords[1] +
                                    load_s_colors(vii[2]) * pick_coords[2] );
                $$ endif
                $$ if color_buffer_channels == 1
                    varyings.color = vec4<f32>(cvalue, cvalue, cvalue, 1.0);
                $$ elif color_buffer_channels == 2
                    varyings.color = vec4<f32>(cvalue.r, cvalue.r, cvalue.r, cvalue.g);
                $$ elif color_buffer_channels == 3
                    varyings.color = vec4<f32>(cvalue, 1.0);
                $$ elif color_buffer_channels == 4
                    varyings.color = vec4<f32>(cvalue);
                $$ endif
            $$ endif

            // Set texture coords
            $$ if color_mode == 'face_map' or color_mode == 'vertex_map'
                $$ if color_mode == 'face_map'
                    let cvalue = load_s_texcoords(face_index);
                $$ else
                    let cvalue = (  load_s_texcoords(vii[0]) * pick_coords[0] +
                                    load_s_texcoords(vii[1]) * pick_coords[1] +
                                    load_s_texcoords(vii[2]) * pick_coords[2] );
                $$ endif
                $$ if colormap_dim == '1d'
                    varyings.texcoord = f32(cvalue);
                $$ elif colormap_dim == '2d'
                    varyings.texcoord = vec2<f32>(cvalue);
                $$ elif colormap_dim == '3d'
                    varyings.texcoord = vec3<f32>(cvalue);
                $$ endif
            $$ endif

            return varyings;
        }
        """

    def code_fragment(self):
        return """
        @fragment
        fn fs_main(varyings: Varyings) -> FragmentOutput {
            // Discart fragments that are too far from the centerline. This makes round caps.
            // Note that we operate in physical pixels here.
            let distx = max(0.0, abs(varyings.dist2center.x) - 0.5 * varyings.segment_length);
            let dist = length(vec2<f32>(distx, varyings.dist2center.y));
            if (dist > varyings.segment_width * 0.5) {
                discard;
            }

            $$ if color_mode == 'vertex' or color_mode == 'face'
                let color_value = varyings.color;
                let albeido = color_value.rgb;
            $$ elif color_mode == 'vertex_map' or color_mode == 'face_map'
                let color_value = sample_colormap(varyings.texcoord);
                let albeido = color_value.rgb;  // no more colormap
            $$ else
                let color_value = u_material.color;
                let albeido = color_value.rgb;
            $$ endif

            // No aa. This is something we need to decide on. See line renderer.
            // Making this < 1 would affect the suggested_render_mask.
            let alpha = 1.0;
            // Set color
            let physical_color = srgb2physical(albeido);
            let opacity = min(1.0, color_value.a) * alpha;
            let out_color = vec4<f32>(physical_color, opacity);

            // Wrap up
            apply_clipping_planes(varyings.world_pos);
            var out = get_fragment_output(varyings.position.z, out_color);
            $$ if write_pick
            // The wobject-id must be 20 bits. In total it must not exceed 64 bits.
            out.pick = (
                pick_pack(u32(u_wobject.id), 20) +
                pick_pack(varyings.pick_idx, 26) +
                pick_pack(u32(varyings.pick_coords.x * 63.0), 6) +
                pick_pack(u32(varyings.pick_coords.y * 63.0), 6) +
                pick_pack(u32(varyings.pick_coords.z * 63.0), 6)
            );
            $$ endif

            // We curve the line away to the background, so that line pieces overlap each-other
            // in a better way, especially avoiding the joins from overlapping the next line piece.
            out.depth = varyings.position.z + 0.0001 * dist;
            return out;
        }
        """



"""
Utilities to load scenes from files, using trimesh.
"""

import pygfx as gfx


def load_scene(path):
    raise DeprecationWarning(
        "The load_scene() function is replaced with load_meshes()."
    )


def load_mesh(path):
    """Load a mesh from a file.

    Parameters
    ----------
    path : str
        The location where the mesh is stored.

    Returns
    -------
    mesh : Mesh
        A pygfx Mesh object. If the file does not contain exactly one mesh, an error is raised.
    """
    meshes = load_meshes(path)
    if len(meshes) != 1:
        raise ValueError(
            f"Found {len(meshes)} meshes instead of 1 in '{path}'. Use `load_meshes()` instead. "
        )
    return meshes[0]


def load_meshes(path):
    """Load meshes from a file.

    This function requires the trimesh library.

    Parameters
    ----------
    path : str
        The location where the mesh or scene is stored.

    Returns
    -------
    meshes : list
        A list of loaded meshes.

    """

    import trimesh  # noqa

    scene = trimesh.load(path)
    if isinstance(scene, trimesh.Trimesh):
        m = gfx.Mesh(
            gfx.geometry_from_trimesh(scene),
            gfx.MeshPhongMaterial(),
        )
        return [m]

    elif isinstance(scene, trimesh.Scene):
        for node_name in scene.graph.nodes_geometry:
            transform, geometry_name = scene.graph[node_name]
            current = scene.geometry[geometry_name]
            current.apply_transform(transform)

        return [
            gfx.Mesh(
                gfx.geometry_from_trimesh(m),
                gfx.material_from_trimesh(m.visual.material),
            )
            for m in scene.geometry.values()
        ]
    else:
        raise ValueError(f"Unexpected trimesh data: {scene.__class__.__name__}")



import wgpu  # only for flags/enums

from . import register_wgpu_render_function, WorldObjectShader, Binding, RenderMask
from ...objects import Points
from ...materials import PointsMaterial, GaussianPointsMaterial


@register_wgpu_render_function(Points, PointsMaterial)
class PointsShader(WorldObjectShader):
    # Notes:
    # In WGPU, the pointsize attribute can no longer be larger than 1 because
    # of restriction in some hardware/backend API's. So we use our storage-buffer
    # approach (similar for what we use for lines) to sort of fake a geometry shader.
    # An alternative is to use instancing. Could be worth testing both approaches
    # for performance ...

    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)
        material = wobject.material
        geometry = wobject.geometry

        color_mode = str(material.color_mode).split(".")[-1]
        if color_mode == "auto":
            if material.map is not None:
                self["color_mode"] = "vertex_map"
                self["color_buffer_channels"] = 0
            else:
                self["color_mode"] = "uniform"
                self["color_buffer_channels"] = 0
        elif color_mode == "uniform":
            self["color_mode"] = "uniform"
            self["color_buffer_channels"] = 0
        elif color_mode == "vertex":
            self["color_mode"] = "vertex"
            self["color_buffer_channels"] = nchannels = geometry.colors.data.shape[1]
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        elif color_mode == "vertex_map":
            self["color_mode"] = "vertex_map"
            self["color_buffer_channels"] = 0
            if material.map is None:
                raise ValueError(f"Cannot apply colormap is no material.map is set.")
        else:
            raise RuntimeError(f"Unknown color_mode: '{color_mode}'")

    def get_bindings(self, wobject, shared):
        geometry = wobject.geometry
        material = wobject.material

        rbuffer = "buffer/read_only_storage"
        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
            Binding("s_positions", rbuffer, geometry.positions, "VERTEX"),
        ]

        self["per_vertex_sizes"] = False
        if material.vertex_sizes:
            self["per_vertex_sizes"] = True
            bindings.append(Binding("s_sizes", rbuffer, geometry.sizes, "VERTEX"))

        # Per-vertex color, colormap, or a uniform color?
        if self["color_mode"] == "vertex":
            bindings.append(Binding("s_colors", rbuffer, geometry.colors, "VERTEX"))
        elif self["color_mode"] == "vertex_map":
            bindings.extend(
                self.define_texcoords_and_colormap(
                    material.map, geometry.texcoords, material.map_interpolation
                )
            )

        self["shape"] = "circle"
        if isinstance(material, GaussianPointsMaterial):
            self["shape"] = "gaussian"

        bindings = {i: b for i, b in enumerate(bindings)}
        self.define_bindings(0, bindings)

        return {
            0: bindings,
        }

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_list,
            "cull_mode": wgpu.CullMode.none,
        }

    def get_render_info(self, wobject, shared):
        material = wobject.material

        offset, size = wobject.geometry.positions.draw_range
        offset, size = offset * 6, size * 6

        render_mask = wobject.render_mask
        if not render_mask:
            if material.is_transparent:
                render_mask = RenderMask.transparent
            elif self["color_mode"] == "uniform":
                if material.color_is_transparent:
                    render_mask = RenderMask.transparent
                else:
                    render_mask = RenderMask.all
            else:
                render_mask = RenderMask.all

        return {
            "indices": (size, 1, offset, 0),
            "render_mask": render_mask,
        }

    def get_code(self):
        return (
            self.code_definitions()
            + self.code_common()
            + self.code_vertex()
            + self.code_fragment()
        )

    def code_vertex(self):
        return """

        struct VertexInput {
            @builtin(vertex_index) vertex_index : u32,
        };


        @vertex
        fn vs_main(in: VertexInput) -> Varyings {

            let index = i32(in.vertex_index);
            let i0 = index / 6;
            let sub_index = index % 6;

            let raw_pos = load_s_positions(i0);
            let world_pos = u_wobject.world_transform * vec4<f32>(raw_pos, 1.0);
            let ndc_pos = u_stdinfo.projection_transform * u_stdinfo.cam_transform * world_pos;

            var deltas = array<vec2<f32>, 6>(
                vec2<f32>(-1.0, -1.0),
                vec2<f32>(-1.0,  1.0),
                vec2<f32>( 1.0, -1.0),
                vec2<f32>(-1.0,  1.0),
                vec2<f32>( 1.0, -1.0),
                vec2<f32>( 1.0,  1.0),
            );

            // Need size here in vertex shader too
            $$ if per_vertex_sizes
                let size = load_s_sizes(i0);
            $$ else
                let size = u_material.size;
            $$ endif

            let aa_margin = 1.0;
            let delta_logical = deltas[sub_index] * (size + aa_margin);
            let delta_ndc = delta_logical * (1.0 / u_stdinfo.logical_size);

            var varyings: Varyings;
            varyings.position = vec4<f32>(ndc_pos.xy + delta_ndc * ndc_pos.w, ndc_pos.zw);
            varyings.world_pos = vec3<f32>(world_pos.xyz / world_pos.w);
            varyings.pointcoord = vec2<f32>(delta_logical);
            varyings.size = f32(size);

            // Picking
            varyings.pick_idx = u32(i0);

            // per-vertex or per-face coloring
            $$ if color_mode == 'face' or color_mode == 'vertex'
                let color_index = i0;
                $$ if color_buffer_channels == 1
                    let cvalue = load_s_colors(color_index);
                    varyings.color = vec4<f32>(cvalue, cvalue, cvalue, 1.0);
                $$ elif color_buffer_channels == 2
                    let cvalue = load_s_colors(color_index);
                    varyings.color = vec4<f32>(cvalue.r, cvalue.r, cvalue.r, cvalue.g);
                $$ elif color_buffer_channels == 3
                    varyings.color = vec4<f32>(load_s_colors(color_index), 1.0);
                $$ elif color_buffer_channels == 4
                    varyings.color = vec4<f32>(load_s_colors(color_index));
                $$ endif
            $$ endif

            // How to index into tex-coords
            let tex_coord_index = i0;

            // Set texture coords
            $$ if colormap_dim == '1d'
            varyings.texcoord = f32(load_s_texcoords(tex_coord_index));
            $$ elif colormap_dim == '2d'
            varyings.texcoord = vec2<f32>(load_s_texcoords(tex_coord_index));
            $$ elif colormap_dim == '3d'
            varyings.texcoord = vec3<f32>(load_s_texcoords(tex_coord_index));
            $$ endif

            return varyings;
        }
        """

    def code_fragment(self):
        # Also see See https://github.com/vispy/vispy/blob/master/vispy/visuals/markers.py
        return """

        @fragment
        fn fs_main(varyings: Varyings) -> FragmentOutput {
            var final_color : vec4<f32>;

            let d = length(varyings.pointcoord);
            let aa_width = 1.0;

            $$ if per_vertex_sizes
                let size = varyings.size;
            $$ else
                let size = u_material.size;
            $$ endif

            $$ if color_mode == 'vertex'
                let color = varyings.color;
            $$ elif color_mode == 'map'
                let color = sample_colormap(varyings.texcoord);
            $$ else
                let color = u_material.color;
            $$ endif

            $$ if shape == 'circle'
                if (d <= size - 0.5 * aa_width) {
                    final_color = color;
                } else if (d <= size + 0.5 * aa_width) {
                    let alpha1 = 0.5 + (size - d) / aa_width;
                    let alpha2 = pow(alpha1, 2.0);  // this works better
                    final_color = vec4<f32>(color.rgb, color.a * alpha2);
                } else {
                    discard;
                }
            $$ elif shape == "gaussian"
                if (d <= size) {
                    let sigma = size / 3.0;
                    let t = d / sigma;
                    let a = exp(-0.5 * t * t);
                    final_color = vec4<f32>(color.rgb, color.a * a);
                } else {
                    discard;
                }
            $$ else
                invalid_point_type;
            $$ endif

            let physical_color = srgb2physical(final_color.rgb);
            let opacity = final_color.a * u_material.opacity;
            let out_color = vec4<f32>(physical_color, opacity);

            // Wrap up
            apply_clipping_planes(varyings.world_pos);
            var out = get_fragment_output(varyings.position.z, out_color);

            $$ if write_pick
            // The wobject-id must be 20 bits. In total it must not exceed 64 bits.
            out.pick = (
                pick_pack(u32(u_wobject.id), 20) +
                pick_pack(varyings.pick_idx, 26) +
                pick_pack(u32(varyings.pointcoord.x + 256.0), 9) +
                pick_pack(u32(varyings.pointcoord.y + 256.0), 9)
            );
            $$ endif

            return out;
        }
        """



import random
import weakref
import threading
import enum
from typing import List
import pylinalg as la

import numpy as np

from ..resources import Buffer
from ..utils import array_from_shadertype, logger
from ..utils.trackable import RootTrackable
from ._events import EventTarget
from ..utils.transform import (
    AffineBase,
    AffineTransform,
    RecursiveTransform,
    callback,
)


class IdProvider:
    """Object for internal use to manage world object id's."""

    def __init__(self):
        self._ids_in_use = set([0])
        self._map = weakref.WeakValueDictionary()
        self._lock = threading.RLock()

    def claim_id(self, wobject):
        """Used by wobjects to claim an id."""
        # We don't simply count up, but keep a pool of ids. This is
        # because an application *could* create and discard objects at
        # a high rate, so we want to be able to re-use these ids.
        #
        # Some numbers:
        # * 4_294_967_296 (2**32) max number for u32
        # * 2_147_483_647 (2**31 -1) max number for i32.
        # *    16_777_216 max integer that can be stored exactly in f32
        # *     4_000_000 max integer that survives being passed as a varying (in my tests)
        # *     1_048_575 is ~1M is 2**20 seems like a good max scene objects.
        # *    67_108_864 is ~50M is 2**26 seems like a good max vertex count.
        #                 which leaves 64-20-26=18 bits for any other picking info.

        # Max allowed id, inclusive
        id_max = 1_048_575  # 2*20-1

        # The max number of ids. This is a bit less to avoid choking
        # when there are few free id's left.
        max_items = 1_000_000

        with self._lock:
            if len(self._ids_in_use) >= max_items:
                raise RuntimeError("Max number of objects reached.")
            id = 0
            while id in self._ids_in_use:
                id = random.randint(1, id_max)
            self._ids_in_use.add(id)
            self._map[id] = wobject

        return id

    def release_id(self, wobject, id):
        """Release an id associated with a wobject."""
        with self._lock:
            self._ids_in_use.discard(id)
            self._map.pop(id, None)

    def get_object_from_id(self, id):
        """Return the wobject associated with an id, or None."""
        return self._map.get(id)


id_provider = IdProvider()


class RenderMask(enum.IntFlag):
    auto = 0
    opaque = 1
    transparent = 2
    all = 3


class WorldObject(EventTarget, RootTrackable):
    """Base class for objects.

    This class represents objects in the world, i.e., the scene graph.Each
    WorldObject has geometry to define it's data, and material to define its
    appearance. The object itself is only responsible for defining object
    hierarchies (parent / children) and its position and orientation in the
    world.

    Parameters
    ----------
    geometry : Geometry
        The data defining the shape of the object.
    material : Material
        The data defining the appearence of the object.
    visible : bool
        Whether the object is visible.
    render_order : int
        The render order (when applicable for the renderer's blend mode).
    render_mask : str
        Determines the render passes that the object is rendered in. It's
        recommended to let the renderer decide, using "auto".

    Notes
    -----
    Use :class:`Group` to collect multiple world objects into a single empty
    world object.

    See Also
    --------
    pygfx.utils.transform.AffineBase
        Various getters and setters defined on ``obj.local`` and ``obj.world``.
    pygfx.utils.transform.AffineTransform
        The class used to implement ``obj.local``.
    pygfx.utils.transform.RecursiveTransform
        The class used to implement ``obj.world``.

    """

    _FORWARD_IS_MINUS_Z = False  # Default is +Z (lights and cameras use -Z)

    # The uniform type describes the structured info for this object, which represents
    # every "propery" that a renderer would need to know in order to visualize it.
    # Put larger items first for alignment, also note that host-sharable structs
    # align at power-of-two only, so e.g. vec3 needs padding.
    # todo: rename uniform to info or something?

    uniform_type = dict(
        world_transform="4x4xf4",
        world_transform_inv="4x4xf4",
        id="i4",
    )

    def __init__(
        self,
        geometry=None,
        material=None,
        *,
        visible=True,
        render_order=0,
        render_mask="auto",
    ):
        super().__init__()
        self._parent: weakref.ReferenceType[WorldObject] = None

        #: Subtrees of the scene graph that depend on this object.
        self.children: List[WorldObject] = []

        self.geometry = geometry
        self.material = material

        # Compose complete uniform type
        buffer = Buffer(array_from_shadertype(self.uniform_type))
        buffer.data["world_transform"] = np.eye(4)
        buffer.data["world_transform_inv"] = np.eye(4)

        #: The object's transform expressed in parent space.
        self.local = AffineTransform(is_camera_space=self._FORWARD_IS_MINUS_Z)
        #: The object's transform expressed in world space.
        self.world = RecursiveTransform(
            self.local, is_camera_space=self._FORWARD_IS_MINUS_Z, reference_up=(0, 1, 0)
        )
        self.world.on_update(self._update_uniform_buffers)

        # Set id
        self._id = id_provider.claim_id(self)
        buffer.data["id"] = self._id

        #: The GPU data of this WorldObject.
        self.uniform_buffer = buffer

        # Init visibility and render props
        self.visible = visible
        self.render_order = render_order
        self.render_mask = render_mask
        self.cast_shadow = False
        self.receive_shadow = False

    @callback
    def _update_uniform_buffers(self, transform: AffineBase):
        self.uniform_buffer.data["world_transform"] = transform.matrix.T
        self.uniform_buffer.data["world_transform_inv"] = transform.inverse_matrix.T
        self.uniform_buffer.update_range()

    def __repr__(self):
        return f"<pygfx.{self.__class__.__name__} at {hex(id(self))}>"

    def __del__(self):
        id_provider.release_id(self, self.id)

    @property
    def up(self):
        """
        Relic of old WorldObjects that aliases with the new ``transform.up``
        direction. Prefer `obj.world.reference_up` instead.

        """

        logger.warning(
            "`WorldObject.up` is deprecated. Use `WorldObject.world.reference_up` instead.",
        )

        return self.world.reference_up

    @up.setter
    def up(self, value):
        logger.warning(
            "`WorldObject.up` is deprecated. Use `WorldObject.world.reference_up` instead.",
        )

        self.world.reference_up = np.asarray(value)

    @property
    def id(self):
        """An integer id smaller than 2**31 (read-only)."""
        return self._id

    @property
    def tracker(self):
        return self._root_tracker

    @property
    def visible(self):
        """Wheter is object is rendered or not. Default True."""
        return self._store.visible

    @visible.setter
    def visible(self, visible):
        self._store.visible = bool(visible)

    @property
    def render_order(self):
        """This value allows the default rendering order of scene graph
        objects to be controlled. Default 0. See ``Renderer.sort_objects``
        for details.
        """
        return self._store.render_order

    @render_order.setter
    def render_order(self, value):
        self._store.render_order = float(value)

    @property
    def render_mask(self):
        """Indicates in what render passes to render this object:

        * "auto": try to determine the best approach (default).
        * "opaque": only in the opaque render pass.
        * "transparent": only in the transparent render pass(es).
        * "all": render in both opaque and transparent render passses.

        If "auto" (the default), the renderer attempts to determine
        whether all fragments will be either opaque or all transparent,
        and only apply the needed render passes. If this cannot be
        determined, it falls back to "all".

        Some objects may contain both transparent and opaque fragments,
        and should be rendered in all passes - the object's contribution
        to each pass is determined on a per-fragment basis.

        For clarity, rendering objects in all passes even though they
        are fully opaque/transparent yields correct results and is
        generally the "safest" option. The only cost is performance.
        Rendering transparent fragments in the opaque pass makes them
        invisible. Rendering opaque fragments in the transparent pass
        blends them as if they are transparent with an alpha of 1.
        """
        return self._store.render_mask

    @render_mask.setter
    def render_mask(self, value):
        if value is None:
            self._store.render_mask = RenderMask(0)
        elif isinstance(value, int):
            self._store.render_mask = RenderMask(value)
        elif isinstance(value, str):
            try:
                self._store.render_mask = RenderMask._member_map_[value.lower()]
            except KeyError:
                opts = set(RenderMask._member_names_)
                msg = f"WorldObject.render_mask must be one of {opts} not {value!r}"
                raise ValueError(msg) from None
        else:
            raise TypeError(
                f"WorldObject.render_mask must be int or str, not {type(value)}"
            )

    @property
    def geometry(self):
        """The object's geometry, the data that defines (the shape of) this object."""
        return self._store.geometry

    @geometry.setter
    def geometry(self, geometry):
        self._store.geometry = geometry

    @property
    def material(self):
        """Wheter is object is rendered or not. Default True."""
        return self._store.material

    @material.setter
    def material(self, material):
        self._store.material = material

    @property
    def cast_shadow(self):
        """Whether this object casts shadows, i.e. whether it is rendered into
        a shadow map. Default False."""
        return self._cast_shadow  # does not affect any shaders

    @cast_shadow.setter
    def cast_shadow(self, value):
        self._cast_shadow = bool(value)

    @property
    def receive_shadow(self):
        """Whether this object receives shadows. Default False."""
        return self._store.receive_shadow

    @receive_shadow.setter
    def receive_shadow(self, value):
        self._store.receive_shadow = bool(value)

    @property
    def parent(self) -> "WorldObject":
        """Object's parent in the scene graph (read-only).
        An object can have at most one parent.
        """
        if self._parent is None:
            return None
        else:
            return self._parent()

    def add(self, *objects, before=None, keep_world_matrix=False) -> "WorldObject":
        """Add child objects.

        Any number of objects may be added. Any current parent on an object
        passed in here will be removed, since an object can have at most one
        parent. If ``before`` argument is given, then the items are inserted
        before the given element.

        Parameters
        ----------
        *objects : WorldObject
            The world objects to add as children.
        before : WorldObject
            If not None, insert the objects before this child object.
        keep_world_matrix : bool
            If True, the child will keep it's world transform. It moves in the
            scene graph but will visually remain in the same place. If False,
            the child will keep it's parent transform.

        """

        obj: WorldObject
        for obj in objects:
            if obj.parent is not None:
                obj.parent.remove(obj, keep_world_matrix=keep_world_matrix)

            if before is not None:
                idx = self.children.index(before)
            else:
                idx = len(self.children)

            if keep_world_matrix:
                transform_matrix = obj.world.matrix

            obj._parent = weakref.ref(self)
            obj.world.parent = self.world
            self.children.insert(idx, obj)

            if keep_world_matrix:
                obj.world.matrix = transform_matrix

        return self

    def remove(self, *objects, keep_world_matrix=False):
        """Removes object as child of this object. Any number of objects may be removed."""

        obj: WorldObject
        for obj in objects:
            try:
                self.children.remove(obj)
            except ValueError:
                logger.warning("Attempting to remove object that was not a child.")
                continue
            else:
                obj._reset_parent(keep_world_matrix=keep_world_matrix)

    def clear(self, *, keep_world_matrix=False):
        """Removes all children."""

        for child in self.children:
            child._reset_parent(keep_world_matrix=keep_world_matrix)

        self.children.clear()

    def _reset_parent(self, *, keep_world_matrix=False):
        """Sets the parent to None.

        xref: https://github.com/pygfx/pygfx/pull/482#discussion_r1135670771
        """

        if keep_world_matrix:
            transform_matrix = self.world.matrix

        self._parent = None
        self.world.parent = None

        if keep_world_matrix:
            self.world.matrix = transform_matrix

    def traverse(self, callback, skip_invisible=False):
        """Executes the callback on this object and all descendants.

        If ``skip_invisible`` is given and True, objects whose
        ``visible`` property is False - and their children - are
        skipped. Note that modifying the scene graph inside the callback
        is discouraged.
        """

        for child in self.iter(skip_invisible=skip_invisible):
            callback(child)

    def iter(self, filter_fn=None, skip_invisible=False):
        """Create a generator that iterates over this objects and its children.
        If ``filter_fn`` is given, only objects for which it returns ``True``
        are included.
        """
        if skip_invisible and not self.visible:
            return

        if filter_fn is None:
            yield self
        elif filter_fn(self):
            yield self

        for child in self.children:
            yield from child.iter(filter_fn, skip_invisible)

    def get_bounding_box(self):
        """Axis-aligned bounding box in parent space.

        Returns
        -------
        aabb : ndarray, [2, 3] or None
            An axis-aligned bounding box, or None when the object does
            not take up a particular space.
        """

        # Collect bounding boxes
        aabbs = []
        for child in self.children:
            aabb = child.get_bounding_box()
            if aabb is not None:
                trafo = child.local.matrix
                aabbs.append(la.aabb_transform(aabb, trafo))
        if self.geometry is not None:
            aabb = self.geometry.get_bounding_box()
            if aabb is not None:
                aabbs.append(aabb)

        # Combine
        if aabbs:
            aabbs = np.stack(aabbs)
            final_aabb = np.zeros((2, 3), dtype=float)
            final_aabb[0] = np.min(aabbs[:, 0, :], axis=0)
            final_aabb[1] = np.max(aabbs[:, 1, :], axis=0)
        else:
            final_aabb = None

        return final_aabb

    def get_bounding_sphere(self):
        """Bounding Sphere in parent space.

        Returns
        -------
        bounding_shere : ndarray, [4] or None
            A sphere (x, y, z, radius), or None when the object does
            not take up a particular space.

        """
        aabb = self.get_bounding_box()
        return None if aabb is None else la.aabb_to_sphere(aabb)

    def get_world_bounding_box(self):
        """Axis aligned bounding box in world space.

        Returns
        -------
        aabb : ndarray, [2, 3] or None
            The transformed axis-aligned bounding box, or None when the
            object does not take up a particular space.

        """
        aabb = self.get_bounding_box()
        return None if aabb is None else la.aabb_transform(aabb, self.world.matrix)

    def get_world_bounding_sphere(self):
        """Bounding Sphere in world space.

        Returns
        -------
        bounding_shere : ndarray, [4] or None
            A sphere (x, y, z, radius), or None when the object does
            not take up a particular space.

        """
        aabb = self.get_world_bounding_box()
        return None if aabb is None else la.aabb_to_sphere(aabb)

    def _wgpu_get_pick_info(self, pick_value):
        # In most cases the material handles this.
        return self.material._wgpu_get_pick_info(pick_value)

    def look_at(self, target) -> None:
        """Orient the object so it looks at the given position.

        This sets the object's rotation such that its ``forward`` direction
        points towards ``target`` (given in world space). This rotation takes
        reference_up into account, i.e., the rotation is chosen in such a way that a
        camera looking ``forward`` follows the rotation of a human head looking
        around without tilting the head sideways.

        """

        self.world.forward = target - self.world.position



"""
A global object shared by all renderers.
"""

import wgpu

from ...resources import Resource, Buffer
from ...utils.trackable import Trackable
from ...utils import array_from_shadertype
from ...utils.text import glyph_atlas
from ._utils import gpu_caches


# Definition uniform struct with standard info related to transforms,
# provided to each shader as uniform at slot 0.
# todo: a combined transform would be nice too, for performance
# todo: same for ndc_to_world transform (combined inv transforms)
stdinfo_uniform_type = dict(
    cam_transform="4x4xf4",
    cam_transform_inv="4x4xf4",
    projection_transform="4x4xf4",
    projection_transform_inv="4x4xf4",
    physical_size="2xf4",
    logical_size="2xf4",
)


class Shared(Trackable):
    """An object to store global data to share between multiple wgpu
    renderers. Each renderer updates the data and then passes this down
    to the pipeline containers.

    The renderer instantiates and stores the singleton shared object.
    """

    # Vanilla WGPU does not support interpolating samplers for float32
    # textures, which is sad for e.g. volume rendering. WebGPU specifies
    # the 'float32-filterable' feature for this, which is also sortof
    # exposed by wgpu-native, except that adapters claim to not support
    # this feature at this point. Fortunately, we can enable it via the
    # native-only feature 'texture_adapter_specific_format_features'

    _features = set(["texture_adapter_specific_format_features"])
    _instance = None

    def __init__(self, *, canvas=None):
        super().__init__()

        assert Shared._instance is None
        Shared._instance = self

        # Create adapter and device objects - there should be just one per canvas.
        # Having a global device provides the benefit that we can draw any object
        # anywhere.
        # We could pass the canvas to request_adapter(), so we get an adapter that is
        # at least compatible with the first canvas that a renderer is create for.
        # However, passing the object has been shown to prevent the creation of
        # a canvas (on Linux + wx), so, we never pass it for now.
        self._adapter = wgpu.request_adapter(
            canvas=None, power_preference="high-performance"
        )
        self._device = self.adapter.request_device(
            required_features=list(Shared._features), required_limits={}
        )

        # Create a uniform buffer for std info
        # Stored on _store so if we'd ever swap it out for another buffer,
        # the pipeline automatically update.
        self._store.uniform_buffer = Buffer(array_from_shadertype(stdinfo_uniform_type))
        self._store.uniform_buffer._wgpu_usage |= wgpu.BufferUsage.UNIFORM

        # Init glyph atlas texture
        self._store.glyph_atlas_texture = None
        self._store.glyph_atlas_info_buffer = None
        self.pre_render_hook()

    def pre_render_hook(self):
        """Called by the renderer on the beginning of a draw."""
        tex = glyph_atlas.texture
        if tex is not self._store["glyph_atlas_texture"]:
            self._store.glyph_atlas_texture = tex
        buffer = glyph_atlas.info_buffer
        if buffer is not self._store["glyph_atlas_info_buffer"]:
            self._store.glyph_atlas_info_buffer = buffer

    @classmethod
    def get_instance(cls):
        return cls._instance

    @property
    def adapter(self):
        """The shared WGPU adapter object."""
        return self._adapter

    @property
    def device(self):
        """The shared WGPU device object."""
        return self._device

    @property
    def uniform_buffer(self):
        """The shared uniform buffer in which the renderer puts
        information about the canvas and camera.
        """
        return self._store.uniform_buffer

    @property
    def glyph_atlas_texture(self):
        """The shared glyph atlas (as a texture view and sampler) for objects that want to render text."""
        return self._store.glyph_atlas_texture

    @property
    def glyph_atlas_info_buffer(self):
        """A buffer containing per-glyph metadata (rects and more)."""
        return self._store.glyph_atlas_info_buffer


def enable_wgpu_features(*features):
    """Enable specific features (as strings) on the wgpu device.

    WARNING: enabling features means that your code may not work on all
    devices. The point of wgpu is that it can make a promise that a
    visualization works and looks the same on any device. Using features
    breaks that promise, and may cause your code to not work on e.g.
    mobile devices or certain operating systems.

    This function must be called before the first wgpu device is created.
    In practice this means before the first ``Renderer`` is created.
    It can be called multiple times to enable more features.

    For more information on features:

    * ``wgpu.FeatureName`` for all possible official features.
    * ``renderer.device.adapter.features`` for the features available on the current system.
    * ``renderer.device.features`` for the currently enabled features.
    * https://gpuweb.github.io/gpuweb/#gpufeaturename for the official webgpu features (excl. native features).
    * https://docs.rs/wgpu/latest/wgpu/struct.Features.html for the features and their limitations in wgpu-core.
    """
    if Shared._instance is not None:
        raise RuntimeError(
            "The enable_wgpu_features() function must be called before creating the first renderer."
        )
    Shared._features.update(features)


def get_shared():
    """Get the globally shared instance. Creates it if it does not yet exist.
    This should not be called at the import time of any module.
    Use this to get the global device: `get_shared().device`.
    """
    if Shared._instance is None:
        Shared()
    return Shared._instance


def print_wgpu_report():
    """Print a report on the internal status of WGPU. Can be useful
    in debugging, and for providing details when making a bug report.
    """
    shared = get_shared()
    adapter = shared.adapter
    device = shared.device

    print()
    print("ADAPTER INFO:")
    for key, val in adapter.request_adapter_info().items():
        print(f"{key.rjust(50)}: {val}")

    print()
    print("FEATURES:".ljust(50), "adapter".rjust(10), "device".rjust(10))
    for key in adapter.features:
        device_has_it = "Y" if key in device.features else "-"
        print(f"{key}:".rjust(50), "Y".rjust(10), device_has_it.rjust(10))

    print()
    print("LIMITS:".ljust(50), "adapter".rjust(10), "device".rjust(10))
    for key in adapter.limits.keys():
        val1 = adapter.limits[key]
        val2 = device.limits.get(key, "-")
        print(f"{key}:".rjust(50), str(val1).rjust(10), str(val2).rjust(10))

    print()
    print("CACHES:".ljust(20), "Count".rjust(10), "Hits".rjust(10), "Misses".rjust(10))
    for cache_name, stats in gpu_caches.get_stats().items():
        count, hits, misses = stats
        print(
            f"{cache_name}".rjust(20),
            str(count).rjust(10),
            str(hits).rjust(10),
            str(misses).rjust(10),
        )

    print()
    print("RESOURCES:".ljust(20), "Count".rjust(10))
    for name, count in Resource._resource_counts.items():
        print(
            f"{name}:".rjust(20),
            str(count).rjust(10),
        )

    print()
    wgpu.print_report()



import numpy as np
import wgpu

from ...resources._texture import Texture
from ._utils import GfxTextureView, GpuCache
from ._shared import get_shared


# This cache enables re-using gpu pipelines for calculating mipmaps, so
# that these don't have to be created on each draw, which would make
# things very slow. The number of pipelines in the cache won't be large
# since there are only so many texture formats, but it seems cleaner
# to use our cache mechanism than to store them globally.
MIPMAP_CACHE = GpuCache("mipmap_pipelines")


mipmap_source = """

    @group(0) @binding(0)
    var t_img1 : texture_2d<TYPE>;

    @group(0) @binding(1)
    var t_img2 : texture_storage_2d<FORMAT, write>;

    @compute
    @workgroup_size(1)
    fn c_main(@builtin(global_invocation_id) index2: vec3<u32>) {

        let texSize1 = vec2<f32>(textureDimensions(t_img1));

        // Determine smoothing settings.
        // Sigma 1.0 matches a factor 2 size reduction.
        // support would ideally be 3*sigma, but we can use less to trade performance
        let sigma = 1.0;
        let support = 2;  // 2: 5x5 kernel, 3: 7x7 kernel

        // For the sampling, we work with integer coords. Also use min/max for the edges.
        let ref_index = vec2<f32>(f32(i32(index2.x) * 2), f32(i32(index2.y) * 2));
        let base_index = vec2<i32>(ref_index);
        let min_index = vec2<i32>(0, 0);
        let max_index = vec2<i32>(texSize1 - 1.0);

        // Convolve. Here we apply a Gaussian kernel, the weight is calculated
        // for each pixel individually based on the distance to the ref_index.
        // This means that the textures don't need to align.
        var val: vec4<f32> = vec4<f32>(0.0, 0.0, 0.0, 0.0);
        var weight: f32 = 0.0;
        for (var y:i32 = -support; y <= support; y = y + 1) {
            for (var x:i32 = -support; x <= support; x = x + 1) {
                let step = vec2<i32>(x, y);
                let index = clamp(base_index + step, min_index, max_index);
                let dist = length(ref_index - vec2<f32>(index) - 0.5);
                let t = dist / sigma;
                let w = exp(-0.5 * t * t);
                let pixelVal = vec4<f32>(textureLoad(t_img1, index, 0));
                val = val + pixelVal * w;
                weight = weight + w;
            }
        }
        let value = vec4<TYPE>(val.rgba / weight);
        let outCoord = vec2<i32>(i32(index2.x), i32(index2.y));
        textureStore(t_img2, outCoord, value);
    }
"""


def get_mip_level_count(texture):
    """Get the number of mipmap levels from the texture size."""
    if not isinstance(texture, Texture):
        raise TypeError("Expecting a Texture object.")
    width, height, _ = texture.size
    return int(np.floor(np.log2(max(width, height))) + 1)


def generate_texture_mipmaps(target):
    """Generate mipmaps for the given target. The target can be a
    Texture or GfxTextureView and can be a 2D texture as well as a cube
    texture.
    """

    # If this looks like a cube or stack, generate mipmaps for each individual layer
    if isinstance(target, Texture) and target.dim == 2 and target.size[2] > 1:
        for i in range(target.size[2]):
            generate_texture_mipmaps(GfxTextureView(target, layer_range=(i, i + 1)))
        return

    if isinstance(target, Texture):
        texture = target
        layer = 0
    elif isinstance(target, GfxTextureView):
        view, texture = target, target.texture
        layer = view.layer_range[0]

    generate_mipmaps(texture, layer)


def generate_mipmaps(texture, base_array_layer):
    device = get_shared().device

    pipeline = get_mipmap_pipeline(device, texture)

    command_encoder: "wgpu.GPUCommandEncoder" = device.create_command_encoder()
    bind_group_layout = pipeline.get_bind_group_layout(0)

    dst_size = texture.size[:2]
    prev_view = None

    for i in range(1, texture._wgpu_mip_level_count):
        dst_size = dst_size[0] // 2, dst_size[1] // 2
        bind_group, prev_view = get_bind_group(
            device, texture, bind_group_layout, base_array_layer, i, prev_view
        )

        pass_encoder = command_encoder.begin_compute_pass()
        pass_encoder.set_pipeline(pipeline)
        pass_encoder.set_bind_group(0, bind_group, [], 0, 99)
        pass_encoder.dispatch_workgroups(dst_size[0], dst_size[1])
        pass_encoder.end()

    device.queue.submit([command_encoder.finish()])


def get_bind_group(
    device, texture, bind_group_layout, base_array_layer, mip_level, src_view
):
    # Get the bind group that includes the two subsequent texture views.
    # We cache this on the texture object to avoid re-creating the views
    # and bind-group each time, which results in a significant increase
    # in performance.
    #
    # Note, however, that this does mean we hold on to GPU resources,
    # which may be a bit of a waste if the mipmaps are only created once!
    # So for now we're assuming that mipmaps are generated often.

    key = f"_gfx_bind_group_{base_array_layer}_{mip_level}"
    bind_group = getattr(texture, key, None)
    dst_view = None
    if bind_group is None:
        if src_view is None:
            src_view = texture._wgpu_object.create_view(
                base_mip_level=mip_level - 1,
                mip_level_count=1,
                dimension="2d",
                base_array_layer=base_array_layer,
            )
        dst_view = texture._wgpu_object.create_view(
            base_mip_level=mip_level,
            mip_level_count=1,
            dimension="2d",
            base_array_layer=base_array_layer,
        )
        bind_group = device.create_bind_group(
            layout=bind_group_layout,
            entries=[
                {"binding": 0, "resource": src_view},
                {"binding": 1, "resource": dst_view},
            ],
        )
        setattr(texture, key, bind_group)

    return bind_group, dst_view


def get_mipmap_pipeline(device, texture):
    format = texture._wgpu_object.format
    pipeline = MIPMAP_CACHE.get(format)

    if pipeline is None:
        if format.endswith("srgb"):
            raise RuntimeError("Cannot create mipmaps for srgb textures.")

        type, textype = "f32", "float"
        if format.endswith(("norm", "float")):
            type, textype = "f32", "float"
        elif format.endswith("sint"):
            type, textype = "i32", "sint"
        elif format.endswith("uint"):
            type, textype = "u32", "uint"

        shader = mipmap_source.replace("FORMAT", format).replace("TYPE", type)
        module = device.create_shader_module(code=shader)

        pipeline = device.create_compute_pipeline(
            layout=create_pipeline_layout(device, format, textype),
            compute={
                "module": module,
                "entry_point": "c_main",
            },
        )

        MIPMAP_CACHE.set(format, pipeline)

    # Strore a ref of the pipeline, the cache uses weak refs.
    # I.e. the pipeline for this format is in the cache for as long as
    # any textures that use it are alive.
    texture._wgpu_mipmap_pipeline = pipeline

    return pipeline


def create_pipeline_layout(device, format, textype):
    bind_group_layouts = []

    entries = []
    entries.append(
        {
            "binding": 0,
            "visibility": wgpu.ShaderStage.COMPUTE,
            "texture": {"sample_type": textype},
        }
    )
    entries.append(
        {
            "binding": 1,
            "visibility": wgpu.ShaderStage.COMPUTE,
            "storage_texture": {
                "access": "write-only",
                "view_dimension": "2d",
                "format": format,
            },
        }
    )

    bind_group_layout = device.create_bind_group_layout(entries=entries)
    bind_group_layouts.append(bind_group_layout)
    pipeline_layout = device.create_pipeline_layout(
        bind_group_layouts=bind_group_layouts
    )
    return pipeline_layout



import numpy as np
import pylinalg as la

from ..utils.trackable import Trackable
from ..resources import Resource, Buffer, Texture


class Geometry(Trackable):
    """Generic container for Geometry data.

    Parameters
    ----------
    kwargs : dict
        A dict of attributes to define on the geometry object. Keys must match
        the naming convention described in the implementation details section of
        the :mod:`Geometries module <pygfx.geometries>`. If they don't they will
        become optional attributes. Values must either be `Resources` or
        ArrayLike.

    Example
    -------

    .. code-block:: py

        g = Geometry(positions=[[1, 2], [2, 4], [3, 5], [4, 1]])
        g.positions.data  # numpy array

    """

    def __init__(self, **kwargs):
        super().__init__()

        self._aabb = None
        self._aabb_rev = None
        self._bsphere = None
        self._bsphere_rev = None

        for name, val in kwargs.items():
            # Get resource object
            if isinstance(val, Resource):
                resource = val
            else:
                # Convert literal arrays to numpy arrays (buffers and textures require memoryview compatible data).
                if isinstance(val, list):
                    dtype = "int32" if name == "indices" else "float32"
                    val = np.array(val, dtype=dtype)
                # Create texture or buffer
                if name == "grid":
                    dim = val.ndim
                    if dim > 2 and val.shape[-1] <= 4:
                        dim -= 1  # last array dim is probably (a subset of) rgba
                    resource = Texture(val, dim=dim)
                else:
                    resource = Buffer(val)

            # Checks
            if isinstance(resource, Buffer):
                format = resource.format
                if name == "indices":
                    # Make no assumptions about shape. Shader will need to validate.
                    # For meshes will be Nx3 or Nx4, but other dtypes may support
                    # multidimensional stuff for fancy graphics.
                    pass
                elif name == "positions":
                    if not format.startswith("3x"):
                        raise ValueError("Expected Nx3 data for positions")
                elif name == "normals":
                    if not format.startswith("3x"):
                        raise ValueError("Expected Nx3 data for normals")
                elif name == "texcoords":
                    if not ("x" not in format or format.startswith(("1x", "2x", "3x"))):
                        raise ValueError("Expected Nx1, Nx2 or Nx3 data for texcoords")
                elif name == "colors":
                    if not format.startswith("4x"):
                        raise ValueError("Expected Nx4 data for colors")
                elif name == "sizes":
                    if not ("x" not in format):
                        raise ValueError("Expected array of scalars for sizes")
                else:
                    pass  # Unknown attribute - no checks

            # Store
            setattr(self, name, resource)

    def __setattr__(self, key, new_value):
        if not key.startswith("_"):
            if isinstance(new_value, Trackable) or key in self._store:
                return setattr(self._store, key, new_value)
        object.__setattr__(self, key, new_value)

    def __getattribute__(self, key):
        if not key.startswith("_"):
            if key in self._store:
                return getattr(self._store, key)
        return object.__getattribute__(self, key)

    def __dir__(self):
        x = object.__dir__(self)
        x.extend(dict.keys(self._store))
        return x

    def get_bounding_box(self):
        """Compute the axis-aligned bounding box.

        Computes the aabb based on either positions or the shape of the grid
        buffer. If both are present, the bounding box will be computed based on
        the positions buffer.

        Returns
        -------
        aabb : ndarray, [2, 3] or None
            The axis-aligned bounding box given by the "smallest" (lowest value)
            and "largest" (highest value) corners. Is None when the geometry has
            no finite positions.

        """
        if hasattr(self, "positions"):
            if self._aabb_rev == self.positions.rev:
                return self._aabb
            aabb = None
            # Get positions and check expected shape
            pos = self.positions.data
            if pos.ndim == 2 and pos.shape[1] in (2, 3):
                # Select finite positions
                finite_mask = np.isfinite(pos).all(axis=1)
                if finite_mask.sum() > 0:
                    # Construct aabb
                    pos_finite = pos[finite_mask]
                    aabb = np.array(
                        [pos_finite.min(axis=0), pos_finite.max(axis=0)], np.float32
                    )
                    # If positions contains xy, but not z, assume z=0
                    if aabb.shape[1] == 2:
                        aabb = np.column_stack([aabb, np.zeros((2, 1), np.float32)])
            self._aabb = aabb
            self._aabb_rev = self.positions.rev
            return self._aabb

        elif hasattr(self, "grid"):
            if self._aabb_rev == self.grid.rev:
                return self._aabb
            # account for multi-channel image data
            grid_shape = self.grid.data.shape[: self.grid.dim]
            # create aabb in index/data space
            aabb = np.array([np.zeros_like(grid_shape), grid_shape[::-1]], dtype="f8")
            # convert to local image space by aligning
            # center of voxel index (0, 0, 0) with origin (0, 0, 0)
            aabb -= 0.5
            # ensure coordinates are 3D
            # NOTE: important we do this last, we don't want to apply
            # the -0.5 offset to the z-coordinate of 2D images
            if aabb.shape[1] == 2:
                aabb = np.hstack([aabb, [[0], [0]]])
            self._aabb = aabb
            self._aabb_rev = self.grid.rev
            return self._aabb
        else:
            return None

    def get_bounding_sphere(self):
        """Compute a bounding sphere.

        Uses the geometry's axis-aligned bounding box, to estimate a sphere
        which contains the geometry.

        Returns
        -------
        sphere : ndarray, [4] or None
            A sphere given by it's center and radius. Format: ``(x, y, z, radius)``.
            Is None when the geometry has no finite positions.

        Notes
        -----
        Since the sphere wraps the geometry's bounding box, it typically won't
        be the minimally binding sphere.

        """

        if hasattr(self, "positions"):
            if self._bsphere_rev == self.positions.rev:
                return self._bsphere
            bsphere = None
            # Get positions and check expected shape
            pos = self.positions.data
            if pos.ndim == 2 and pos.shape[1] in (2, 3):
                # Select finite positions
                finite_mask = np.isfinite(pos).all(axis=1)
                if finite_mask.sum() > 0:
                    # Construct aabb
                    pos_finite = pos[finite_mask]
                    center = pos_finite.mean(axis=0)
                    distances = np.linalg.norm(pos_finite - center, axis=0)
                    radius = float(distances.max())
                    if len(center) == 2:
                        bsphere = np.array(
                            [center[0], center[1], 0.0, radius], np.float32
                        )
                    else:
                        bsphere = np.array(
                            [center[0], center[1], center[1], radius], np.float32
                        )
            self._bsphere = bsphere
            self._bsphere_rev = self.positions.rev
            return self._bsphere

        else:
            if self._bsphere_rev == self._aabb_rev:
                return self._bsphere
            aabb = self.get_bounding_box()
            self._bsphere = None if aabb is None else la.aabb_to_sphere(aabb)
            self._bsphere_rev = self._aabb_rev
            return self._bsphere



import re

text_prog = re.compile(r"\s+")


def tokenize_text(text):
    """Splits the text in pieces of "ws" and "other" (whitespace and non-whitespace)."""
    pos = 0
    while True:
        match = text_prog.search(text, pos)

        if not match:
            other = text[pos:]
            if other:
                yield "other", other
            break

        other = text[pos : match.start()]
        if other:
            yield "other", other

        s = match.group()
        yield "ws", s
        pos = match.end()


punctuation = r"[\,\.\;\:\?\!\…]"
markdown_prog = re.compile(r"(\s)+|(\d|\w|-|_)+|(\*)+|(" + punctuation + ")")


def tokenize_markdown(text):
    """Splits the text in pieces of: ws, words, stars, punctuation, other."""
    pos = 0
    while True:
        match = markdown_prog.search(text, pos)

        if not match:
            other = text[pos:]
            if other:
                yield "other", other
            break

        other = text[pos : match.start()]
        if other:
            yield "", other

        s = match.group()

        if match.group(1):
            yield "ws", s
        elif match.group(2):
            yield "word", s
        elif match.group(3):
            yield "stars", s
        elif match.group(4):
            yield "punctuation", s
        else:
            yield "other", s  # should not happen, but just in case

        pos = match.end()



import numpy as np

from ._base import Resource, get_item_format_from_memoryview


class Buffer(Resource):
    """A contiguous piece of GPU memory.

    Buffers can be used as index buffer, vertex buffer, uniform buffer, or
    storage buffer. You can provide (and update data for it), or use it as a
    placeholder for a buffer with no representation on the CPU.

    Parameters
    ----------
    data : array
        The initial data of the array data. It must support the buffer-protocol,
        (e.g. a bytes or numpy array). If None, nbytes and nitems must be
        provided.
    nbytes : int
        The size of the buffer in bytes. Ignored if ``data`` is used.
    nitems : int
        The number of elements in the buffer. Ignored if ``data`` is used.
    format : str
        A format string describing the buffer layout when used as a vertex
        buffer. If None, this is automatically set from the data. This must be a
        pygfx format specifier, e.g. "3xf4", but can also be a format specific
        to the render backend if necessary (e.g. from ``wgpu.VertexFormat``).
    """

    def __init__(
        self,
        data=None,
        *,
        nbytes=None,
        nitems=None,
        format=None,
    ):
        super().__init__()
        Resource._rev += 1
        self._rev = Resource._rev
        # To specify the buffer size
        # The actual data (optional)
        self._data = None
        self._format = None
        self._gfx_pending_uploads = []  # list of (offset, size) tuples

        # Backends-specific attributes for internal use
        self._wgpu_object = None
        self._wgpu_usage = 0

        # Get nbytes
        if data is not None:
            self._data = data
            self._mem = mem = memoryview(data)
            subformat = get_item_format_from_memoryview(mem)
            if subformat:
                shape = (mem.shape + (1,)) if len(mem.shape) == 1 else mem.shape
                if len(shape) == 2:  # if not, the user does something fancy
                    self._format = (f"{shape[-1]}x" + subformat).lstrip("1x")
            the_nbytes = mem.nbytes
            the_nitems = mem.shape[0] if mem.shape else 1
            if the_nitems:
                self._gfx_pending_uploads.append((0, the_nitems))
            if nbytes is not None and nbytes != the_nbytes:
                raise ValueError("Given nbytes does not match size of given data.")
            if nitems is not None and nitems != the_nitems:
                raise ValueError("Given nitems does not match shape of given data.")
        elif nbytes is not None and nitems is not None:
            the_nbytes = int(nbytes)
            the_nitems = int(nitems)
        else:
            raise ValueError(
                "Buffer must be instantiated with either data or nbytes and nitems."
            )

        if format is not None:
            self._format = str(format)

        self._store.nbytes = the_nbytes
        self._store.nitems = the_nitems

        self.draw_range = 0, the_nitems

    @property
    def rev(self):
        """An integer that is increased when update_range() is called."""
        return self._rev

    @property
    def data(self):
        """The data for this buffer. Can be None if the data only
        exists on the GPU.

        Note: the data is the same reference that was given to instantiate this
        object, but this may change.
        """
        return self._data

    @property
    def mem(self):
        """The data for this buffer as a memoryview. Can be None if
        the data only exists on the GPU.
        """
        return self._mem

    @property
    def nbytes(self):
        """The number of bytes in the buffer."""
        return self._store.nbytes

    @property
    def nitems(self):
        """The number of items in the buffer."""
        return self._store.nitems

    @property
    def itemsize(self):
        """The number of bytes for a single item."""
        if self._data is None:
            # This generic solution fails when nitems is zero
            return self._store.nbytes // self._store.nitems
        else:
            shape = self._mem.shape
            if shape:
                shape = shape[1:]
            factor = int(np.prod(shape)) or 1
            return factor * self._mem.itemsize

    @property
    def format(self):
        """The buffer format.

        Usually a pygfx format specifier (e.g. u2 for scalar uint16,
        or 3xf4 for 3xfloat32), but can also be a overriden to a
        backend-specific format. Can also be None e.g. for uniform buffers.
        """
        return self._format

    @property
    def vertex_byte_range(self):
        raise DeprecationWarning(
            "vertex_byte_range is deprecated, use draw_range instead."
        )

    @vertex_byte_range.setter
    def vertex_byte_range(self, offset_nbytes):
        raise DeprecationWarning(
            "vertex_byte_range is deprecated, use draw_range instead."
        )

    @property
    def draw_range(self):
        """The range to data (origin, size) expressed in items."""
        return self._store.draw_range

    @draw_range.setter
    def draw_range(self, draw_range):
        origin, size = draw_range
        origin, size = int(origin), int(size)
        if not (origin == 0 or 0 < origin < self.nitems):  # note nitems can be 0
            raise ValueError("draw_range origin out of bounds.")
        if not (size >= 0 and origin + size <= self.nitems):
            raise ValueError("draw_range size out of bounds.")
        self._store.draw_range = origin, size
        Resource._rev += 1
        self._rev = Resource._rev

    def update_range(self, offset=0, size=2**50):
        """Mark a certain range of the data for upload to the GPU. The
        offset and size are expressed in integer number of elements.
        """
        # See ThreeJS BufferAttribute.updateRange
        # Check input
        if not isinstance(offset, int) and isinstance(size, int):
            raise TypeError(
                f"`offset` and `size` must be native `int` type, you have passed: "
                f"offset: <{type(offset)}>, size: <{type(size)}>"
            )
        offset, size = int(offset), int(size)  # convert np ints to real ints

        if size == 0:
            return
        elif size < 0:
            raise ValueError("Update size must not be negative")
        elif offset < 0:
            raise ValueError("Update offset must not be negative")
        elif offset + size > self.nitems:
            size = self.nitems - offset
        # Merge with current entry?
        if self._gfx_pending_uploads:
            cur_offset, cur_size = self._gfx_pending_uploads.pop(-1)
            end = max(offset + size, cur_offset + cur_size)
            offset = min(offset, cur_offset)
            size = end - offset
        # Limit and apply
        self._gfx_pending_uploads.append((offset, size))
        Resource._rev += 1
        self._rev = Resource._rev
        self._gfx_mark_for_sync()
        # note: this can be smarter, we have logic for chunking in the morph tool

    def _get_subdata(self, offset, size):
        """Return subdata as a contiguous array."""
        # If this is a full range, this is easy
        if offset == 0 and size == self.nitems and self.mem.contiguous:
            return self.mem
        # Get a numpy array, because memoryviews do not support nd slicing
        if isinstance(self.data, np.ndarray):
            arr = self.data
        elif not self.mem.c_contiguous:
            raise ValueError(
                "Non-contiguous texture data is only supported for numpy array."
            )
        else:
            arr = np.frombuffer(self.mem, self.mem.format).reshape(self.mem.shape)
        # Slice it
        sub_arr = arr[offset : offset + size]
        return memoryview(np.ascontiguousarray(sub_arr))



from ._base import Material, ColorMode
from ..resources import Texture
from ..utils import unpack_bitfield, Color


class PointsMaterial(Material):
    """Point default material.

    Renders (antialiased) disks of the given size and color.

    Parameters
    ----------
    color : Color
        The uniform color of the points (used depending on the ``color_mode``).
    size : int
        The size (diameter) of the points in screen space (px). Ignored if
        vertex_size is True.
    color_mode : enum or str
        The mode by which the line is coloured. Default 'auto'.
    vertex_sizes : bool
        If True, use the vertex sizes provided in the geometry to set point
        sizes.
    map : Texture
        The texture map specifying the color for each texture coordinate.
    map_interpolation: str
        The method to interpolate the color map. Either 'nearest' or 'linear'. Default 'linear'.
    kwargs : Any
        Additional kwargs will be passed to the :class:`material base class
        <pygfx.Material>`.

    """

    uniform_type = dict(
        Material.uniform_type,
        color="4xf4",
        size="f4",
    )

    def __init__(
        self,
        color=(1, 1, 1, 1),
        size=1,
        color_mode="auto",
        vertex_sizes=False,
        map=None,
        map_interpolation="linear",
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.color = color
        self.color_mode = color_mode
        self.map = map
        self.map_interpolation = map_interpolation
        self.size = size
        self._vertex_sizes = bool(vertex_sizes)

    def _wgpu_get_pick_info(self, pick_value):
        # This should match with the shader
        values = unpack_bitfield(pick_value, wobject_id=20, index=26, x=9, y=9)
        return {
            "vertex_index": values["index"],
            "point_coord": (values["x"] - 256.0, values["y"] - 256.0),
        }

    @property
    def color(self):
        """The color of the points (if map is not set)."""
        return Color(self.uniform_buffer.data["color"])

    @color.setter
    def color(self, color):
        color = Color(color)
        self.uniform_buffer.data["color"] = color
        self.uniform_buffer.update_range(0, 1)
        self._store.color_is_transparent = color.a < 1

    @property
    def color_is_transparent(self):
        """Whether the color is (semi) transparent (i.e. not fully opaque)."""
        return self._store.color_is_transparent

    @property
    def color_mode(self):
        """The way that color is applied to the mesh.

        * auto: switch between `uniform` and `vertex_map`, depending on whether `map` is set.
        * uniform: use the material's color property for the whole mesh.
        * vertex: use the geomerty `colors` buffer, one color per vertex.
        * vertex_map: use the geometry texcoords buffer to sample (per vertex) in the material's ``map`` texture.
        """
        return self._store.color_mode

    @color_mode.setter
    def color_mode(self, value):
        if isinstance(value, ColorMode):
            pass
        elif isinstance(value, str):
            if value.startswith("ColorMode."):
                value = value.split(".")[-1]
            try:
                value = getattr(ColorMode, value.lower())
            except AttributeError:
                raise ValueError(f"Invalid color_mode: '{value}'")
        else:
            raise TypeError(f"Invalid color_mode class: {value.__class__.__name__}")
        if value == ColorMode.face or value == ColorMode.face_map:
            raise ValueError(f"Points cannot have color_mode {value}")
        self._store.color_mode = value

    @property
    def vertex_colors(self):
        return self.color_mode == ColorMode.vertex

    @vertex_colors.setter
    def vertex_colors(self, value):
        raise DeprecationWarning(
            "vertex_colors is deprecated, use ``color_mode='vertex'``"
        )

    @property
    def size(self):
        """The size (diameter) of the points, in logical pixels."""
        return float(self.uniform_buffer.data["size"])

    @size.setter
    def size(self, size):
        self.uniform_buffer.data["size"] = size
        self.uniform_buffer.update_range(0, 1)

    @property
    def vertex_sizes(self):
        """Whether to use the vertex sizes provided in the geometry."""
        return self._vertex_sizes

    @vertex_sizes.setter
    def vertex_sizes(self, value):
        value = bool(value)
        if value != self._vertex_sizes:
            self._vertex_sizes = value

    @property
    def map(self):
        """The texture map specifying the color for each texture coordinate.
        The dimensionality of the map can be 1D, 2D or 3D, but should match the
        number of columns in the geometry's texcoords.
        """
        return self._map

    @map.setter
    def map(self, map):
        assert map is None or isinstance(map, Texture)
        self._map = map

    @property
    def map_interpolation(self):
        """The method to interpolate the colormap. Either 'nearest' or 'linear'."""
        return self._store.map_interpolation

    @map_interpolation.setter
    def map_interpolation(self, value):
        assert value in ("nearest", "linear")
        self._store.map_interpolation = value

    # todo: sizeAttenuation


class GaussianPointsMaterial(PointsMaterial):
    """A material for points, renders Gaussian blobs with a standard
    deviation of 1/6 of the size.
    """


# idea: a MarkerMaterial with more options for the shape, and an edge around the shape.
# Though perhaps such a material should be part of a higher level plotting lib.



from ..resources import Texture
from ._base import Material


class VolumeBasicMaterial(Material):
    """Basic volume material.

    Parameters
    ----------
    clim : tuple
        The contrast limits to scale the data values with. Default (0, 1).
    map : Texture
        The colormap to turn the voxel values into their final color.
    interpolation : str
        The method to interpolate the image data. Either 'nearest' or 'linear'. Default 'linear'.
    map_interpolation: str
        The method to interpolate the color map. Either 'nearest' or 'linear'. Default 'linear'.
    kwargs : Any
        Additional kwargs will be passed to the :class:`material base class
        <pygfx.Material>`.

    """

    uniform_type = dict(
        Material.uniform_type,
        clim="2xf4",
    )

    def __init__(
        self,
        clim=None,
        map=None,
        interpolation="linear",
        map_interpolation="linear",
        **kwargs
    ):
        super().__init__(**kwargs)
        self.map = map
        self.clim = clim
        # Note: the default volume interpolation is 'linear' while it's nearest
        # for images. The ability to spot the individual voxels simply results in
        # poor visual quality.
        self.interpolation = interpolation
        self.map_interpolation = map_interpolation

    @property
    def map(self):
        """The colormap to turn the voxel values into their final color.
        If not given or None, the values themselves represents the color.
        The dimensionality of the map can be 1D, 2D or 3D, but should match the
        number of channels in the volume.
        """
        return self._map

    @map.setter
    def map(self, map):
        assert map is None or isinstance(map, Texture)
        self._map = map

    @property
    def clim(self):
        """The contrast limits to scale the data values with. Default (0, 1)."""
        v1, v2 = self.uniform_buffer.data["clim"]
        return float(v1), float(v2)

    @clim.setter
    def clim(self, clim):
        # Check and store given clim
        if clim is None:
            clim = 0, 1
        clim = float(clim[0]), float(clim[1])
        # Update uniform data
        self.uniform_buffer.data["clim"] = clim
        self.uniform_buffer.update_range(0, 1)

    @property
    def interpolation(self):
        """The method to interpolate the image data. Either 'nearest' or 'linear'."""
        return self._store.interpolation

    @interpolation.setter
    def interpolation(self, value):
        assert value in ("nearest", "linear")
        self._store.interpolation = value

    @property
    def map_interpolation(self):
        """The method to interpolate the colormap. Either 'nearest' or 'linear'."""
        return self._store.map_interpolation

    @map_interpolation.setter
    def map_interpolation(self, value):
        assert value in ("nearest", "linear")
        self._store.map_interpolation = value


class VolumeSliceMaterial(VolumeBasicMaterial):
    """A material for rendering a slice through a 3D texture at the surface of a mesh.
    This material is not affected by lights.
    """

    uniform_type = dict(
        VolumeBasicMaterial.uniform_type,
        plane="4xf4",
    )

    def __init__(self, plane=(0, 0, 1, 0), **kwargs):
        super().__init__(**kwargs)
        self.plane = plane

    @property
    def plane(self):
        """The plane to slice at, represented with 4 floats ``(a, b, c, d)``,
        which make up the equation: ``ax + by + cz + d = 0`` The plane
        definition applies to the world space (of the scene).
        """
        return self.uniform_buffer.data["plane"]

    @plane.setter
    def plane(self, plane):
        self.uniform_buffer.data["plane"] = plane
        self.uniform_buffer.update_range(0, 1)


class VolumeRayMaterial(VolumeBasicMaterial):
    """A material for rendering volumes using raycasting."""

    # todo: define render modes as material subclasses or using a `mode` or `style` property?
    render_mode = "mip"


class VolumeMipMaterial(VolumeRayMaterial):
    """A material rendering a volume using MIP rendering."""


class VolumeIsoMaterial(VolumeRayMaterial):
    """A material rendering a volume using isosurface rendering."""



from ._base import Material, ColorMode
from ..resources import Texture
from ..utils import unpack_bitfield, Color


class LineMaterial(Material):
    """Basic line material.

    Parameters
    ----------
    color : Color
        The uniform color of the line (used depending on the ``color_mode``).
    thickness : float
        The line thickness expressed in logical pixels.
    color_mode : enum or str
        The mode by which the line is coloured. Default 'auto'.
    map : Texture
        The texture map specifying the color for each texture coordinate. Opional.
    map_interpolation: str
        The method to interpolate the color map. Either 'nearest' or 'linear'. Default 'linear'.
    aa : bool
        Whether or not the line should be anti-aliased. Aliasing gives prettier
        results, but may affect performance for very large datasets. Default
        True.
    kwargs : Any
        Additional kwargs will be passed to the :class:`material base class
        <pygfx.Material>`.

    """

    uniform_type = dict(
        Material.uniform_type,
        color="4xf4",
        thickness="f4",
    )

    def __init__(
        self,
        color=(1, 1, 1, 1),
        thickness=2.0,
        color_mode="auto",
        map=None,
        map_interpolation="linear",
        aa=True,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.color = color
        self.aa = aa
        self.map = map
        self.map_interpolation = map_interpolation
        self.thickness = thickness
        self.color_mode = color_mode

    def _wgpu_get_pick_info(self, pick_value):
        # This should match with the shader
        values = unpack_bitfield(pick_value, wobject_id=20, index=26, coord=18)
        return {
            "vertex_index": values["index"],
            "segment_coord": (values["coord"] - 100000) / 100000.0,
        }

    @property
    def color(self):
        """The uniform color of the line."""
        return Color(self.uniform_buffer.data["color"])

    @color.setter
    def color(self, color):
        color = Color(color)
        self.uniform_buffer.data["color"] = color
        self.uniform_buffer.update_range(0, 1)
        self._store.color_is_transparent = color.a < 1

    @property
    def color_is_transparent(self):
        """Whether the color is (semi) transparent (i.e. not fully opaque)."""
        return self._store.color_is_transparent

    @property
    def aa(self):
        """Whether or not the line should be anti-aliased. Aliasing
        gives prettier results, but may affect performance for very large
        datasets. Default True.
        """
        return self._store.aa

    @aa.setter
    def aa(self, aa):
        self._store.aa = bool(aa)

    @property
    def color_mode(self):
        """The way that color is applied to the mesh.

        * auto: switch between `uniform` and `vertex_map`, depending on whether `map` is set.
        * uniform: use the material's color property for the whole mesh.
        * vertex: use the geomerty `colors` buffer, one color per vertex.
        * face: use the geomerty `colors` buffer, one color per line-piece.
        * vertex_map: use the geometry texcoords buffer to sample (per vertex) in the material's ``map`` texture.
        * faces_map: use the geometry texcoords buffer to sample (per line-piece) in the material's ``map`` texture.
        """
        return self._store.color_mode

    @color_mode.setter
    def color_mode(self, value):
        if isinstance(value, ColorMode):
            pass
        elif isinstance(value, str):
            if value.startswith("ColorMode."):
                value = value.split(".")[-1]
            try:
                value = getattr(ColorMode, value.lower())
            except AttributeError:
                raise ValueError(f"Invalid color_mode: '{value}'")
        else:
            raise TypeError(f"Invalid color_mode class: {value.__class__.__name__}")
        self._store.color_mode = value

    @property
    def vertex_colors(self):
        return self.color_mode == ColorMode.vertex

    @vertex_colors.setter
    def vertex_colors(self, value):
        raise DeprecationWarning(
            "vertex_colors is deprecated, use ``color_mode='vertex'``"
        )

    @property
    def thickness(self):
        """The line thickness expressed in logical pixels."""
        return float(self.uniform_buffer.data["thickness"])

    @thickness.setter
    def thickness(self, thickness):
        self.uniform_buffer.data["thickness"] = thickness
        self.uniform_buffer.update_range(0, 1)

    @property
    def map(self):
        """The texture map specifying the color for each texture coordinate.
        Can be None. The dimensionality of the map can be 1D, 2D or 3D,
        but should match the number of columns in the geometry's
        texcoords.
        """
        return self._map

    @map.setter
    def map(self, map):
        assert map is None or isinstance(map, Texture)
        self._map = map

    @property
    def map_interpolation(self):
        """The method to interpolate the colormap. Either 'nearest' or 'linear'."""
        return self._store.map_interpolation

    @map_interpolation.setter
    def map_interpolation(self, value):
        assert value in ("nearest", "linear")
        self._store.map_interpolation = value


class LineThinMaterial(LineMaterial):
    """Thin line material.

    A simple line, drawn with line_strip primitives that has a thickness
    of one physical pixel (the thickness property is ignored).

    While you probably don't want to use this property in your application (its
    width is inconsistent and looks *very* thin on HiDPI monitors), it can be
    useful for debugging as it is more performant than other line materials.

    """


class LineThinSegmentMaterial(LineMaterial):
    """Thin line segment material.

    Simple line segments, drawn with line primitives that has a thickness
    of one physical pixel (the thickness property is ignored).

    While you probably don't want to use this property in your application (its
    width is inconsistent and looks *very* thin on HiDPI monitors), it can be
    useful for debugging as it is more performant than other line materials.

    """


class LineSegmentMaterial(LineMaterial):
    """Line segment material.

    A material that renders line segments between each two subsequent points."""


class LineArrowMaterial(LineSegmentMaterial):
    """Arrow (vector) line material.

    A material that renders line segments that look like little vectors."""



import sys
import subprocess

import wgpu
import pytest
import pygfx as gfx

from .testutils import can_use_wgpu_lib


def test_shared_object_not_created_at_import():
    code = "import pygfx; print(pygfx.renderers.wgpu._shared.Shared._instance)"
    p = subprocess.run(
        [
            sys.executable,
            "-c",
            code,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
    )
    result = p.stdout.strip()

    print(result)
    assert result.endswith("None")
    # If this fails, either the test is broken, or pygfx calles get_shared()
    # at import-time somewhere, which we don't want to do.


def test_shared_object():
    if not can_use_wgpu_lib:
        pytest.skip("Skipping tests that need the wgpu lib", allow_module_level=True)

    get_shared = gfx.renderers.wgpu._shared.get_shared

    assert get_shared() is not None
    assert isinstance(get_shared().adapter, wgpu.GPUAdapter)
    assert isinstance(get_shared().device, wgpu.GPUDevice)

    # Check that multiple invokations produce the same object - its a singleton
    assert get_shared() is get_shared()



def test_pyi_pygfx(pyi_builder):
    pyi_builder.test_source(
        """
        import pygfx as gfx
        ff = gfx.font_manager.select_font("foo", gfx.font_manager.default_font_props)[0][1]
        print(ff.codepoints)
        assert len(ff.codepoints) > 100
        """
    )



from ._mesh import MeshStandardMaterial
from ..utils.color import Color
from ..resources import Texture


def texture_from_pillow_image(image, dim=2, **kwargs):
    """Pillow Image texture.

    Create a Texture from a PIL.Image.

    Parameters
    ----------
    image : Image
        The `PIL.Image
        <https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image>`_
        to convert into a texture.
    dim : int
        The number of spatial dimensions of the image.
    kwargs : Any
        Additional kwargs are forwarded to :class:`pygfx.Texture`.

    Returns
    -------
    image_texture : Texture
        A texture object representing the given image.

    """

    from PIL.Image import Image  # noqa

    if not isinstance(image, Image):
        raise NotImplementedError()

    m = memoryview(image.tobytes())

    im_channels = len(image.getbands())
    buffer_shape = image.size + (im_channels,)

    m = m.cast(m.format, shape=buffer_shape)
    return Texture(m, dim=dim, **kwargs)


def material_from_trimesh(material):
    """Convert a Trimesh material object into a pygfx material.

    Parameters
    ----------
    material : trimesh.Material
        The material to convert.

    Returns
    -------
    converted : Material
        The converted material.

    """
    from trimesh.visual.material import PBRMaterial  # noqa

    if not isinstance(material, PBRMaterial):
        raise NotImplementedError()

    gfx_material = MeshStandardMaterial()

    if material.baseColorTexture is not None:
        gfx_material.map = texture_from_pillow_image(material.baseColorTexture)

    if material.emissiveFactor is not None:
        gfx_material.emissive = Color(*material.emissiveFactor)

    if material.emissiveTexture is not None:
        gfx_material.emissive_map = texture_from_pillow_image(material.emissiveTexture)

    if material.metallicRoughnessTexture is not None:
        metallic_roughness_map = texture_from_pillow_image(
            material.metallicRoughnessTexture
        )
        gfx_material.roughness_map = metallic_roughness_map
        gfx_material.metalness_map = metallic_roughness_map
        gfx_material.roughness = 1.0
        gfx_material.metalness = 1.0

    if material.roughnessFactor is not None:
        gfx_material.roughness = material.roughnessFactor

    if material.metallicFactor is not None:
        gfx_material.metalness = material.metallicFactor

    if material.normalTexture is not None:
        gfx_material.normal_map = texture_from_pillow_image(material.normalTexture)
        # See: https://github.com/KhronosGroup/glTF-Sample-Models/tree/master/2.0/NormalTangentTest#problem-flipped-y-axis-or-flipped-green-channel
        gfx_material.normal_scale = (1.0, -1.0)

    if material.occlusionTexture is not None:
        gfx_material.ao_map = texture_from_pillow_image(material.occlusionTexture)

    gfx_material.side = "FRONT"
    return gfx_material
